import numpy as np
from numpy.testing import assert_allclose

from light_curve.light_curve_py import MedianAbsoluteDeviation


def test_medabsdev():
    m = [1.0, 1.0, 3.0, 4.0, 7.0]
    feature = MedianAbsoluteDeviation()
    actual = feature(np.linspace(0, 1, len(m)), m, None)
    desired = 2.0
    assert_allclose(actual, desired)



from abc import abstractmethod
from dataclasses import dataclass
from typing import Dict, List, Tuple

import numpy as np

from light_curve.light_curve_py.dataclass_field import dataclass_field
from light_curve.light_curve_py.features._base import BaseMultiBandFeature
from light_curve.light_curve_py.features.rainbow._bands import Bands
from light_curve.light_curve_py.features.rainbow._parameters import create_parameters_class
from light_curve.light_curve_py.features.rainbow._scaler import MultiBandScaler, Scaler
from light_curve.light_curve_py.minuit_lsq import LeastSquares
from light_curve.light_curve_py.minuit_ml import MaximumLikelihood

__all__ = ["BaseRainbowFit"]


# CODATA 2018, grab from astropy
planck_constant = 6.62607004e-27  # erg s
speed_of_light = 2.99792458e10  # cm/s
boltzman_constant = 1.380649e-16  # erg/K
sigma_sb = 5.6703744191844314e-05  # erg/(cm^2 s K^4)


IMINUIT_IMPORT_ERROR = (
    "The `iminuit` package v2.21.0 or larger (exists for Python >= 3.8 only) is required for RainbowFit, "
    "please install it manually or reinstall light-curve package with [full] extra"
)


@dataclass()
class BaseRainbowFit(BaseMultiBandFeature):
    band_wave_cm: Dict[str, float]
    """Mapping of band names and their effective wavelengths in cm."""

    with_baseline: bool = dataclass_field(default=True, kw_only=True)
    """Whether to include a baseline in the fit, one per band."""

    fail_on_divergence: bool = dataclass_field(default=True, kw_only=True)
    """Fail (or return [fill_value]*n) if optimization hasn't converged"""

    @property
    def is_band_required(self) -> bool:
        return True

    @property
    def is_multiband_supported(self) -> bool:
        return True

    @property
    def size(self) -> int:
        return len(self.p)

    @staticmethod
    @abstractmethod
    def _common_parameter_names() -> List[str]:
        """Common parameter names."""
        return NotImplementedError

    @staticmethod
    @abstractmethod
    def _bolometric_parameter_names() -> List[str]:
        """Bolometric parameter names."""
        return NotImplementedError

    @staticmethod
    @abstractmethod
    def _temperature_parameter_names() -> List[str]:
        """Temperature parameter names."""
        return NotImplementedError

    def __post_init__(self) -> None:
        super().__post_init__()

        self._initialize_minuit()

        self.bands = Bands.from_dict(self.band_wave_cm)

        self.p = create_parameters_class(
            f"{self.__class__.__name__}Parameters",
            common=self._common_parameter_names(),
            bol=self._bolometric_parameter_names(),
            temp=self._temperature_parameter_names(),
            bands=self.bands,
            with_baseline=self.with_baseline,
        )

        if len(self.band_wave_cm) == 0:
            raise ValueError("At least one band must be specified.")

        # We are going to use it for the normalization of amplitude,
        # it is normally should be by the order of (\sigma_SB T^4) / (\pi B_\nu)
        self.average_nu = speed_of_light / self.bands.mean_wave_cm

        if self.with_baseline:
            self._lsq_model = self._lsq_model_with_baseline
        else:
            self._lsq_model = self._lsq_model_no_baseline

    def _initialize_minuit(self) -> None:
        self._check_iminuit()

        from iminuit import Minuit

        self.Minuit = Minuit

    @classmethod
    def from_nm(cls, band_wave_nm, **kwargs):
        """Initialize from a dictionary of band names and their effective wavelengths in nm."""
        band_wave_nm = {band: 1e-7 * wavelength for band, wavelength in band_wave_nm.items()}
        return cls(band_wave_cm=band_wave_nm, **kwargs)

    @classmethod
    def from_angstrom(cls, band_wave_aa, **kwargs):
        """Initialize from a dictionary of band names and their effective wavelengths in angstroms."""
        band_wave_aa = {band: 1e-8 * wavelength for band, wavelength in band_wave_aa.items()}
        return cls(band_wave_cm=band_wave_aa, **kwargs)

    @staticmethod
    def _check_iminuit():
        if LeastSquares is None:
            raise ImportError(IMINUIT_IMPORT_ERROR)

        if MaximumLikelihood is None:
            raise ImportError(IMINUIT_IMPORT_ERROR)

        try:
            try:
                from packaging.version import parse as parse_version
            except ImportError:
                from distutils.version import LooseVersion as parse_version

            from iminuit import __version__
        except ImportError:
            raise ImportError(IMINUIT_IMPORT_ERROR)

        if parse_version(__version__) < parse_version("2.21.0"):
            raise ImportError(IMINUIT_IMPORT_ERROR)

    @abstractmethod
    def bol_func(self, t, params):
        """Bolometric light curve function."""
        raise NotImplementedError

    @abstractmethod
    def temp_func(self, t, params):
        """Temperature evolution function."""
        return NotImplementedError

    def _parameter_scalings(self) -> Dict[str, str]:
        """Rules for scaling/unscaling the parameters"""
        rules = {}

        if self.with_baseline:
            for band_name in self.bands.names:
                baseline_name = self.p.baseline_parameter_name(band_name)
                rules[baseline_name] = "baseline"

        return rules

    def _parameter_scale(self, name: str, t_scaler: Scaler, m_scaler: MultiBandScaler) -> float:
        """Return the scale factor to be applied to the parameter to unscale it"""
        scaling = self._parameter_scalings().get(name)
        if scaling == "time" or scaling == "timescale":
            return t_scaler.scale
        elif scaling == "flux":
            return m_scaler.scale

        return 1

    def _unscale_parameters(self, params, t_scaler: Scaler, m_scaler: MultiBandScaler) -> None:
        """Unscale parameters from internal units, in-place."""
        for name, scaling in self._parameter_scalings().items():
            if scaling == "time":
                params[self.p[name]] = t_scaler.undo_shift_scale(params[self.p[name]])

            elif scaling == "timescale":
                params[self.p[name]] = t_scaler.undo_scale(params[self.p[name]])

            elif scaling == "flux":
                params[self.p[name]] = m_scaler.undo_scale(params[self.p[name]])

            elif scaling == "baseline":
                band_name = self.p.baseline_band_name(name)
                baseline = params[self.p[name]]
                params[self.p[name]] = m_scaler.undo_shift_scale_band(baseline, band_name)

                pass

            elif scaling is None or scaling.lower() == "none":
                pass

            else:
                raise ValueError("Unsupported parameter scaling: " + scaling)

    def _unscale_errors(self, errors, t_scaler: Scaler, m_scaler: MultiBandScaler) -> None:
        """Unscale parameter errors from internal units, in-place."""
        for name in self.names:
            scale = self._parameter_scale(name, t_scaler, m_scaler)
            errors[self.p[name]] *= scale

    def _unscale_covariance(self, cov, t_scaler: Scaler, m_scaler: MultiBandScaler) -> None:
        """Unscale parameter covariance from internal units, in-place."""
        for name in self.names:
            scale = self._parameter_scale(name, t_scaler, m_scaler)
            i = self.p[name]
            cov[:, i] *= scale
            cov[i, :] *= scale

    @staticmethod
    def planck_nu(wave_cm, T):
        """Planck function in frequency units."""
        nu = speed_of_light / wave_cm
        return (
            (2 * planck_constant / speed_of_light**2) * nu**3 / np.expm1(planck_constant * nu / (boltzman_constant * T))
        )

    def _lsq_model_no_baseline(self, x, *params):
        """Model function for the fit."""
        t, _band_idx, wave_cm = x
        params = np.array(params)

        bol = self.bol_func(t, params)
        temp = self.temp_func(t, params)

        # Normalize the Planck function so that the result is of order unity
        norm = (sigma_sb * temp**4) / np.pi / self.average_nu  # Original "bolometric" normalization
        # peak_nu =  2.821 * boltzman_constant * temp / planck_constant # Wien displacement law
        # norm = self.planck_nu(speed_of_light / peak_nu, temp) # Peak = 1 normalization

        planck = self.planck_nu(wave_cm, temp) / norm

        flux = planck * bol

        return flux

    def _lsq_model_with_baseline(self, x, *params):
        flux = self._lsq_model_no_baseline(x, *params)

        _t, band_idx, _wave_cm = x
        params = np.array(params)
        baselines = params[self.p.lookup_baseline_idx_with_band_idx(band_idx)]
        flux += baselines

        return flux

    def model(self, t, band, *params):
        """Model function for the fit."""
        band_idx = self.bands.get_index(band)
        wave_cm = self.bands.index_to_wave_cm(band_idx)
        params = np.array(params)
        return self._lsq_model((t, band_idx, wave_cm), *params)

    @property
    def names(self):
        """Names of the parameters."""
        return list(self.p.__members__)

    @abstractmethod
    def _initial_guesses(self, t, m, sigma, band) -> Dict[str, float]:
        """Initial guesses for the fit parameters.

        t and m are *scaled* arrays. No baseline parameters are included.
        """
        return NotImplementedError

    def _baseline_initial_guesses(self, t, m, sigma, band) -> Dict[str, float]:
        """Initial guesses for the baseline parameters."""
        del t
        return {
            self.p.baseline_parameter_name(b): (np.median(m[band == b]) if np.sum(band == b) else 0)
            for b in self.bands.names
        }

    @abstractmethod
    def _limits(self, t, m, sigma, band) -> Dict[str, Tuple[float, float]]:
        """Limits for the fit parameters.

        t and m are *scaled* arrays. No baseline parameters are included.
        """
        return NotImplementedError

    def _baseline_limits(self, t, m, sigma, band) -> Dict[str, Tuple[float, float]]:
        """Limits for the baseline parameters."""
        del t
        limits = {}
        for b in self.bands.names:
            m_band = m[band == b]
            if len(m_band) > 0:
                lower = np.min(m_band) - 10 * np.ptp(m_band)
                upper = np.max(m_band)
            else:
                lower = 0
                upper = 0
            limits[self.p.baseline_parameter_name(b)] = (lower, upper)
        return limits

    def _eval(self, *, t, m, sigma, band):
        params, errors = self._eval_and_get_errors(t=t, m=m, sigma=sigma, band=band)
        return params

    # This is abstractmethod, but we could use default implementation while _eval is defined
    def _eval_and_fill(self, *, t, m, sigma, band, fill_value):
        return super()._eval_and_fill(t=t, m=m, sigma=sigma, band=band, fill_value=fill_value)

    def _eval_and_get_errors(
        self,
        *,
        t,
        m,
        sigma,
        band,
        upper_mask=None,
        get_initial=False,
        return_covariance=False,
        print_level=None,
        debug=False,
    ):
        # Initialize data scalers
        t_scaler = Scaler.from_time(t)
        m_scaler = MultiBandScaler.from_flux(m, band, with_baseline=self.with_baseline)

        # normalize input data
        t = t_scaler.do_shift_scale(t)
        m = m_scaler.do_shift_scale(m)
        sigma = m_scaler.do_scale(sigma)

        band_idx = self.bands.get_index(band)
        wave_cm = self.bands.index_to_wave_cm(band_idx)

        if self.with_baseline:
            initial_baselines = self._baseline_initial_guesses(t, m, sigma, band)
            m_corr = m - np.array([initial_baselines[self.p.baseline_parameter_name(b)] for b in band])

            # Compute initial guesses for the parameters on baseline-subtracted data
            initial_guesses = self._initial_guesses(t, m_corr, sigma, band)
            limits = self._limits(t, m_corr, sigma, band)

            initial_guesses.update(initial_baselines)
            limits.update(self._baseline_limits(t, m, sigma, band))
        else:
            # Compute initial guesses for the parameters on original data
            initial_guesses = self._initial_guesses(t, m, sigma, band)
            limits = self._limits(t, m, sigma, band)

        # least_squares = LeastSquares(
        cost_function = MaximumLikelihood(
            model=self._lsq_model,
            parameters=limits,
            x=(t, band_idx, wave_cm),
            y=m,
            yerror=sigma,
            upper_mask=upper_mask,
        )
        minuit = self.Minuit(cost_function, name=self.names, **initial_guesses)
        # TODO: expose these parameters through function arguments
        if print_level is not None:
            minuit.print_level = print_level
        minuit.strategy = 0  # We will need to manually call .hesse() on convergence anyway

        # Supposedly it is not the same as just setting iterate=10?..
        for i in range(10):
            minuit.migrad()

            if minuit.valid:
                minuit.hesse()
                # hesse() may may drive it invalid
                if minuit.valid:
                    break
            else:
                # That's what iterate is supposed to do?..
                minuit.simplex()
                # FIXME: it may drive the fit valid, but we will not have Hesse run on last iteration

        if debug:
            # Expose everything we have to outside, unscaled, for easier debugging
            self.minuit = minuit
            self.mparams = {
                "t": t,
                "band_idx": band_idx,
                "wave_cm": wave_cm,
                "m": m,
                "sigma": sigma,
                "limits": limits,
                "upper_mask": upper_mask,
                "initial_guesses": initial_guesses,
                "values": minuit.values,
                "errors": minuit.errors,
                "covariance": minuit.covariance,
            }

        if not minuit.valid and self.fail_on_divergence and not get_initial:
            raise RuntimeError("Fitting failed")

        reduced_chi2 = minuit.fval / (len(t) - self.size)

        if get_initial:
            # Reset the fitter so that it returns initial values instead of final ones
            minuit.reset()

        params = np.array(minuit.values)
        errors = np.array(minuit.errors)

        self._unscale_parameters(params, t_scaler, m_scaler)

        # Unscale errors
        self._unscale_errors(errors, t_scaler, m_scaler)

        return_values = np.r_[params, reduced_chi2], errors

        if return_covariance:
            # Unscale covaiance
            cov = np.array(minuit.covariance)
            self._unscale_covariance(cov, t_scaler, m_scaler)
            return_values += (cov,)

        return return_values

    def fit_and_get_errors(self, t, m, sigma, band, *, sorted=None, check=True, **kwargs):
        t, m, sigma, band = self._normalize_input(t=t, m=m, sigma=sigma, band=band, sorted=sorted, check=check)

        return self._eval_and_get_errors(t=t, m=m, sigma=sigma, band=band, **kwargs)



import numpy as np

from ._base import BaseSingleBandFeature


class Amplitude(BaseSingleBandFeature):
    def _eval_single_band(self, t, m, sigma=None):
        return 0.5 * np.ptp(m)

    @property
    def size_single_band(self):
        return 1


__all__ = ("Amplitude",)



from dataclasses import dataclass
from typing import Dict, List, Tuple, Union

from light_curve.light_curve_py.dataclass_field import dataclass_field
from light_curve.light_curve_py.features.rainbow._base import BaseRainbowFit

from .bolometric import BaseBolometricTerm, bolometric_terms
from .temperature import BaseTemperatureTerm, temperature_terms

__all__ = ["RainbowFit"]

# CODATA 2018, grab from astropy
planck_constant = 6.62607004e-27  # erg s
speed_of_light = 2.99792458e10  # cm/s
boltzman_constant = 1.380649e-16  # erg/K
sigma_sb = 5.6703744191844314e-05  # erg/(cm^2 s K^4)


@dataclass()
class RainbowFit(BaseRainbowFit):
    """Multiband blackbody fit to the light curve using functions to be chosen by the user
    Note, that `m` and corresponded `sigma` are assumed to be flux densities.
    Based on Russeil et al. 2023, arXiv:2310.02916.

    Parameters
    ----------
    band_wave_cm : dict
        Dictionary of band names and their effective wavelengths in cm.
    with_baseline : bool, optional
        Whether to include an offset in the fit, individual for each band.
        If it is true, one more fit paramter per passband is added -
        the additive constant with the same units as input flux.
    bolometric : str or BaseBolometricTerm subclass, optional
        The shape of bolometric term. Default is 'bazin'.
        Other options are: 'sigmoid'
    temperature : str or BaseTemperatureTerm subclass, optional
        The shape of temperature term. Default is 'sigmoid'.
        Other options are: 'constant', 'delayed_sigmoid'

    Methods
    -------
    __call__(t, m, sigma, band, *, sorted=False, check=True, fill_value=None)
        Evaluate the feature. Positional arguments are numpy arrays of the same length,
        `band` must consist of the same strings as keys in `band_wave_cm`. If `sorted` is True,
        `t` must be sorted in ascending order. If `check` is True, the input is checked for
        NaNs and Infs. If `fill_value` is not None, it is used to fill the output array if
        the feature cannot be evaluated.
    model(t, band, *params)
        Evaluate Rainbow model on the given arrays of times and bands. `*params` are
        fit parameters, basically the output of `__call__` method but without the last
        parameter (reduced Chi^2 of the fit). See parameter names in the `.name` attribute.
    peak_time(*params)
        Return bolometric peak time for given set of parameters
    fit_and_get_errors(t, m, sigma, band, sorted=False, check=True, print_level=None, get_initial=False)
        The same as `__call__` but also returns the parameter errors. Optionally sets the `print_level`
        (verbosity) for Minuit fitter. If `get_initial` is True, returns the initial parameters instead
        of fitted ones (useful for debugging)
    """

    bolometric: Union[str, BaseBolometricTerm] = dataclass_field(default="bazin", kw_only=True)
    """Which parametric bolometric term to use"""
    temperature: Union[str, BaseTemperatureTerm] = dataclass_field(default="sigmoid", kw_only=True)
    """Which parametric temperature term to use"""

    def __post_init__(self):
        if not isinstance(self.bolometric, BaseBolometricTerm):
            self.bolometric = bolometric_terms[self.bolometric]

        if not isinstance(self.temperature, BaseTemperatureTerm):
            self.temperature = temperature_terms[self.temperature]

        super().__post_init__()

    def _common_parameter_names(self) -> List[str]:
        bolometric_parameters = self.bolometric.parameter_names()
        temperature_parameters = self.temperature.parameter_names()
        return [j for j in bolometric_parameters if j in temperature_parameters]

    def _bolometric_parameter_names(self) -> List[str]:
        bolometric_parameters = self.bolometric.parameter_names()
        return [i for i in bolometric_parameters if i not in self._common_parameter_names()]

    def _temperature_parameter_names(self) -> List[str]:
        temperature_parameters = self.temperature.parameter_names()
        return [i for i in temperature_parameters if i not in self._common_parameter_names()]

    def bol_func(self, t, params):
        return self.bolometric.value(t, *params[self.p.all_bol_idx])

    def temp_func(self, t, params):
        return self.temperature.value(t, *params[self.p.all_temp_idx])

    def _parameter_scalings(self) -> Dict[str, str]:
        rules = super()._parameter_scalings()

        for term in [self.bolometric, self.temperature]:
            for name, scaling in zip(term.parameter_names(), term.parameter_scalings()):
                rules[name] = scaling

        return rules

    def _initial_guesses(self, t, m, sigma, band) -> Dict[str, float]:
        initial = self.bolometric.initial_guesses(t, m, sigma, band)
        initial.update(self.temperature.initial_guesses(t, m, sigma, band))

        return initial

    def _limits(self, t, m, sigma, band) -> Dict[str, Tuple[float, float]]:
        limits = self.bolometric.limits(t, m, sigma, band)
        limits.update(self.temperature.limits(t, m, sigma, band))

        return limits

    def peak_time(self, params) -> float:
        """Returns true bolometric peak position for given parameters"""
        return self.bolometric.peak_time(*params[self.p.all_bol_idx])



import numpy as np
import pytest

from light_curve.light_curve_py import LinearFit


def test_no_init_bands_no_input_bands():
    """Announce no bands, use no bands"""
    t = [1.0, 2.0, 3.0, 4.0, 5.0]
    m = [1.0, 2.0, 3.0, 4.0, 5.0]
    sigma = [1.0, 1.0, 1.0, 1.0, 1.0]
    feature = LinearFit(bands=None)
    _values = feature(t, m, sigma)


def test_init_bands_no_input_bands():
    """Announce bands, use no bands"""
    t = [1.0, 2.0, 3.0, 4.0, 5.0]
    m = [1.0, 2.0, 3.0, 4.0, 5.0]
    sigma = [1.0, 1.0, 1.0, 1.0, 1.0]
    feature = LinearFit(bands=["g"])
    with pytest.raises(ValueError):
        _values = feature(t, m, sigma)


def test_no_init_bands_input_bands():
    """Announce no bands, use input bands"""
    t = [1.0, 2.0, 3.0, 4.0, 5.0]
    m = [1.0, 2.0, 3.0, 4.0, 5.0]
    sigma = [1.0, 1.0, 1.0, 1.0, 1.0]
    bands = ["g", "g", "g", "r", "r"]
    feature = LinearFit(bands=None)
    with pytest.raises(ValueError):
        _values = feature(t, m, sigma, bands)


def test_init_bands_eq_input_bands():
    """Announce no bands, use input bands"""
    t = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0]
    m = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0]
    sigma = [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
    bands = ["g", "g", "g", "r", "r", "r"]
    feature = LinearFit(bands=["g", "r"])
    values = feature(t, m, sigma, bands)
    assert 2 == feature.n_bands
    assert values.size == feature.size
    assert values.size == feature.size_single_band * feature.n_bands


def test_init_bands_less_input_bands():
    t = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0]
    m = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0]
    sigma = [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
    bands = ["g", "g", "g", "r", "r", "r"]
    feature = LinearFit(bands=["g"])
    values = feature(t, m, sigma, bands)
    assert 1 == feature.n_bands
    assert values.size == feature.size
    assert values.size == feature.size_single_band


def test_init_bands_more_input_bands():
    t = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0]
    m = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0]
    sigma = [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
    bands = ["g"] * 6
    feature = LinearFit(bands=["g", "r"])
    # "r" band is virtually existing, but having zero observations
    with pytest.raises(ValueError):
        _values = feature(t, m, sigma, bands, fill_value=None)
    # ... so it should be filled with the fill_value
    values = feature(t, m, sigma, bands, fill_value=np.nan)
    assert 2 == feature.n_bands
    assert values.size == feature.size
    assert values.size == feature.size_single_band * feature.n_bands
    assert np.all(np.isfinite(values[: feature.size_single_band]))
    assert np.all(np.isnan(values[feature.size_single_band :]))


def test_init_bands_different_input_bands():
    t = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0]
    m = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0]
    sigma = [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
    bands = ["g"] * 6
    feature = LinearFit(bands=["r"])
    # "r" band is virtually existing, but having zero observations
    with pytest.raises(ValueError):
        _values = feature(t, m, sigma, bands, fill_value=None)
    # ... so it should be filled with the fill_value
    values = feature(t, m, sigma, bands, fill_value=-999.0)
    assert 1 == feature.n_bands
    assert values.size == feature.size == feature.size_single_band
    np.testing.assert_array_equal(values, -999.0)



import numpy as np

from ._base import BaseSingleBandFeature


class PeakToPeakVar(BaseSingleBandFeature):
    r"""Peak-to-peak variation

    $$
    \frac{(m_i - \sigma_i)_\text{max} - (m_i + \sigma_i)_\text{min}}
    {(m_i - \sigma_i)_\text{max} + (m_i + \sigma_i)_\text{min}}
    $$
    Input m must be non-negative (e.g. non-differential) flux density.
    This feature is a variability detector, higher values correspond to more variable sources.

    - Depends on: **flux density**, **errors**
    - Minimum number of observations: **2**
    - Number of features: **1**

    Aller M.F., Aller H.D., Hughes P.A. 1992. [DOI:10.1086/171898](https://www.doi.org/10.1086/171898)
    """

    nstd: float = 1.0

    def _eval_single_band(self, t, m, sigma=None):
        if np.any(m < 0):
            raise ValueError("m must be non-negative")
        a = np.max(m - self.nstd * sigma)
        b = np.min(m + self.nstd * sigma)
        return (a - b) / (a + b)

    @property
    def size_single_band(self):
        return 1


__all__ = ("PeakToPeakVar",)



import numpy as np

from ._base import BaseSingleBandFeature
from ._lstsq import least_squares


class LinearTrend(BaseSingleBandFeature):
    def _eval_single_band(self, t, m, sigma=None):
        n = len(t)

        slope, chi2 = least_squares(t, m, None)

        red_chi2 = chi2 / (n - 2)
        sxx = np.var(t, ddof=n - 1)
        return np.array([slope, np.sqrt(red_chi2 / sxx), np.sqrt(red_chi2)])

    @property
    def size_single_band(self):
        return 3


__all__ = ("LinearTrend",)



from .features.adnormal import *
from .features.amplitude import *
from .features.beyondnstd import *
from .features.bins import *
from .features.color_of_median import *
from .features.cusum import *
from .features.eta import *
from .features.etae import *
from .features.excvar import *
from .features.extractor import *
from .features.flux_n_not_det_before_fd import *
from .features.intpercrange import *
from .features.kurtosis import *
from .features.linfit import *
from .features.lintrend import *
from .features.magnitude_n_not_det_before_fd import *
from .features.magnpratio import *
from .features.maxslope import *
from .features.mean import *
from .features.meanvar import *
from .features.medabsdev import *
from .features.medbufrperc import *
from .features.median import *
from .features.otsusplit import *
from .features.pdiffmperc import *
from .features.percampl import *
from .features.ptp_var import *
from .features.rainbow import *
from .features.redchi2 import *
from .features.roms import *
from .features.skew import *
from .features.stdev import *
from .features.stetsonk import *
from .features.weightmean import *



from numpy.testing import assert_allclose

from light_curve.light_curve_py import BeyondNStd


def test_beyondnstd_1():
    m = [1.0, 2.0, 3.0, 4.0, 5.0, 10.0]
    feature = BeyondNStd(nstd=1.0)
    feature_2 = BeyondNStd()
    actual = feature(m, m)
    actual_2 = feature_2(m, m)
    desired = 1 / 6
    assert_allclose(actual, desired, actual_2)


def test_beyondnsts_2():
    m = [1.0, 18.0, 45.0, 100.0]
    feature = BeyondNStd(nstd=3.0)
    actual = feature(m, m)
    desired = 0.0
    assert_allclose(actual, desired)



import math
from abc import abstractmethod
from dataclasses import dataclass
from typing import Dict, List, Union

import numpy as np

__all__ = [
    "bolometric_terms",
    "BaseBolometricTerm",
    "SigmoidBolometricTerm",
    "BazinBolometricTerm",
    "LinexpBolometricTerm",
    "DoublexpBolometricTerm",
]


@dataclass()
class BaseBolometricTerm:
    """Bolometric term for the Rainbow"""

    @staticmethod
    @abstractmethod
    def parameter_names() -> List[str]:
        return NotImplementedError

    @staticmethod
    @abstractmethod
    def parameter_scalings() -> List[Union[str, None]]:
        """Describes how to unscale the parameters.

        Should be the list the same shape as returned by `parameter_names()`, and describes
        how the parameters should be un-scaled from the fit done in scaled coordinates.

        List items should be either None or one of the following strings:
        - time - the parameter is scaled and shifted like measurement times
        - timescale - the parameter is scaled like measurement times, but not shifted, thus
            behaving like a difference between two measurement times
        - flux - the parameter is scaled like the flux points, without additional shifts
            applied to them. Suitable for amplitude-like parameters.
        - None - the parameter is kept as is, without any additional scaling or shifting
        """
        return NotImplementedError

    @staticmethod
    @abstractmethod
    def value(self, t, *params) -> float:
        return NotImplementedError

    @staticmethod
    @abstractmethod
    def initial_guesses(t, m, sigma, band) -> Dict[str, float]:
        return NotImplementedError

    @staticmethod
    @abstractmethod
    def limits(t, m, sigma, band) -> Dict[str, float]:
        return NotImplementedError

    @staticmethod
    @abstractmethod
    def peak_time(*params) -> float:
        return NotImplementedError


@dataclass()
class SigmoidBolometricTerm(BaseBolometricTerm):
    """Sigmoid"""

    @staticmethod
    def parameter_names():
        return ["reference_time", "amplitude", "rise_time"]

    @staticmethod
    def parameter_scalings():
        return ["time", "flux", "timescale"]

    @staticmethod
    def value(t, t0, amplitude, rise_time):
        dt = t - t0

        result = np.zeros(len(dt))
        # To avoid numerical overflows, let's only compute the exponents not too far from t0
        idx = dt > -100 * rise_time
        result[idx] = amplitude / (np.exp(-dt[idx] / rise_time) + 1)

        return result

    @staticmethod
    def initial_guesses(t, m, sigma, band):
        A = np.ptp(m)

        initial = {}
        initial["reference_time"] = t[np.argmax(m)]
        initial["amplitude"] = A
        initial["rise_time"] = 1.0

        return initial

    @staticmethod
    def limits(t, m, sigma, band):
        t_amplitude = np.ptp(t)
        m_amplitude = np.ptp(m)

        mean_dt = np.median(t[1:] - t[:-1])

        limits = {}
        limits["reference_time"] = (np.min(t) - 10 * t_amplitude, np.max(t) + 10 * t_amplitude)
        limits["amplitude"] = (0.0, 20 * m_amplitude)
        limits["rise_time"] = (0.1 * mean_dt, 10 * t_amplitude)

        return limits

    @staticmethod
    def peak_time(t0, amplitude, rise_time):
        """Peak time is not defined for the sigmoid, so it returns mid-time of the rise instead"""
        return t0


@dataclass()
class BazinBolometricTerm(BaseBolometricTerm):
    """Bazin function, symmetric form"""

    @staticmethod
    def parameter_names():
        return ["reference_time", "amplitude", "rise_time", "fall_time"]

    @staticmethod
    def parameter_scalings():
        return ["time", "flux", "timescale", "timescale"]

    @staticmethod
    def value(t, t0, amplitude, rise_time, fall_time):
        dt = t - t0

        # Coefficient to make peak amplitude equal to unity
        scale = (fall_time / rise_time) ** (rise_time / (fall_time + rise_time)) + (fall_time / rise_time) ** (
            -fall_time / (fall_time + rise_time)
        )

        result = np.zeros(len(dt))
        # To avoid numerical overflows, let's only compute the exponents not too far from t0
        idx = (dt > -100 * rise_time) & (dt < 100 * fall_time)
        result[idx] = amplitude * scale / (np.exp(-dt[idx] / rise_time) + np.exp(dt[idx] / fall_time))

        return result

    @staticmethod
    def initial_guesses(t, m, sigma, band):
        A = np.ptp(m)

        mc = m - np.min(m)  # To avoid crashing on all-negative data

        # Naive peak position from the highest point
        t0 = t[np.argmax(m)]
        # Peak position as weighted centroid of everything above median
        idx = m > np.median(m)
        # t0 = np.sum(t[idx] * m[idx] / sigma[idx]) / np.sum(m[idx] / sigma[idx])
        # Weighted centroid sigma
        dt = np.sqrt(np.sum((t[idx] - t0) ** 2 * (mc[idx]) / sigma[idx]) / np.sum(mc[idx] / sigma[idx]))

        # Empirical conversion of sigma to rise/fall times
        rise_time = dt / 2
        fall_time = dt / 2

        # Compensate for the difference between reference_time and peak position
        t0 -= np.log(fall_time / rise_time) * rise_time * fall_time / (rise_time + fall_time)

        initial = {}
        initial["reference_time"] = t0
        initial["amplitude"] = A
        initial["rise_time"] = rise_time
        initial["fall_time"] = fall_time

        return initial

    @staticmethod
    def limits(t, m, sigma, band):
        t_amplitude = np.ptp(t)
        m_amplitude = np.ptp(m)

        mean_dt = np.median(t[1:] - t[:-1])

        limits = {}
        limits["reference_time"] = (np.min(t) - 10 * t_amplitude, np.max(t) + 10 * t_amplitude)
        limits["amplitude"] = (0.0, 20 * m_amplitude)
        limits["rise_time"] = (0.1 * mean_dt, 10 * t_amplitude)
        limits["fall_time"] = (0.1 * mean_dt, 10 * t_amplitude)

        return limits

    @staticmethod
    def peak_time(t0, amplitude, rise_time, fall_time):
        return t0 + np.log(fall_time / rise_time) * rise_time * fall_time / (rise_time + fall_time)


@dataclass()
class LinexpBolometricTerm(BaseBolometricTerm):
    """Linexp function, symmetric form. Generated using a prototype version of Multi-view
    Symbolic Regression (Russeil et al. 2024, https://arxiv.org/abs/2402.04298) on
    a SLSN ZTF light curve (https://ztf.snad.space/dr17/view/821207100004043)"""

    @staticmethod
    def parameter_names():
        return ["reference_time", "amplitude", "rise_time"]

    @staticmethod
    def parameter_scalings():
        return ["time", "flux", "timescale"]

    @staticmethod
    def value(t, t0, amplitude, rise_time):
        dt = t0 - t
        protected_rise = math.copysign(max(1e-5, abs(rise_time)), rise_time)

        # Coefficient to make peak amplitude equal to unity
        scale = 1 / (protected_rise * np.exp(-1))

        power = -dt / protected_rise
        power = np.where(power > 100, 100, power)
        result = amplitude * scale * dt * np.exp(power)

        return result

    @staticmethod
    def initial_guesses(t, m, sigma, band):
        A = np.ptp(m)
        med_dt = median_dt(t, band)

        # Compute points after or before maximum
        peak_time = t[np.argmax(m)]
        after = t[-1] - peak_time
        before = peak_time - t[0]

        rise_time = 100 * med_dt
        rise_time = rise_time if before >= after else -rise_time

        initial = {}
        # Reference of linexp correspond to the moment where flux == 0
        initial["reference_time"] = peak_time + rise_time
        initial["amplitude"] = A
        initial["rise_time"] = rise_time

        return initial

    @staticmethod
    def limits(t, m, sigma, band):
        t_amplitude = np.ptp(t)
        m_amplitude = np.ptp(m)

        limits = {}
        limits["reference_time"] = (np.min(t) - 10 * t_amplitude, np.max(t) + 10 * t_amplitude)
        limits["amplitude"] = (0, 10 * m_amplitude)
        limits["rise_time"] = (-10 * t_amplitude, 10 * t_amplitude)

        return limits

    @staticmethod
    def peak_time(t0, amplitude, rise_time):
        return t0 - rise_time


@dataclass()
class DoublexpBolometricTerm(BaseBolometricTerm):
    """Doublexp function generated using Multi-view Symbolic Regression on ZTF SNIa light curves
    Russeil et al. 2024, https://arxiv.org/abs/2402.04298"""

    @staticmethod
    def parameter_names():
        return ["reference_time", "amplitude", "time1", "time2", "p"]

    @staticmethod
    def parameter_scalings():
        return ["time", "flux", "timescale", "timescale", "None"]

    @staticmethod
    def value(t, t0, amplitude, time1, time2, p):
        dt = t - t0

        result = np.zeros_like(dt)

        # To avoid numerical overflows
        maxp = 20
        A = -(dt / time1) * (p - np.exp(-(dt / time2)))
        A = np.where(A > maxp, maxp, A)

        result = amplitude * np.exp(A)

        return result

    @staticmethod
    def initial_guesses(t, m, sigma, band):
        A = np.ptp(m)
        med_dt = median_dt(t, band)

        # Naive peak position from the highest point
        t0 = t[np.argmax(m)]

        # Empirical conversion of sigma to rise/fall times
        time1 = 50 * med_dt
        time2 = 50 * med_dt

        initial = {}
        initial["reference_time"] = t0
        initial["amplitude"] = A
        initial["time1"] = time1
        initial["time2"] = time2
        initial["p"] = 0.1

        return initial

    @staticmethod
    def limits(t, m, sigma, band):
        t_amplitude = np.ptp(t)
        m_amplitude = np.ptp(m)
        med_dt = median_dt(t, band)

        limits = {}
        limits["reference_time"] = (np.min(t) - 10 * t_amplitude, np.max(t) + 10 * t_amplitude)
        limits["amplitude"] = (0.0, 10 * m_amplitude)
        limits["time1"] = (med_dt, 2 * t_amplitude)
        limits["time2"] = (med_dt, 2 * t_amplitude)
        limits["p"] = (1e-4, 10)

        return limits

    @staticmethod
    def peak_time(t0, p):
        try:
            from scipy.special import lambertw
        except ImportError:
            raise ImportError("scipy is required for DoublexpBolometricTerm.peak_time, please install it")

        return t0 + np.real(-lambertw(p * np.exp(1)) + 1)


def median_dt(t, band):
    # Compute the median distance between points in each band
    dt = []
    for b in np.unique(band):
        dt += list(t[band == b][1:] - t[band == b][:-1])
    med_dt = np.median(dt)
    return med_dt


bolometric_terms = {
    "sigmoid": SigmoidBolometricTerm,
    "bazin": BazinBolometricTerm,
    "linexp": LinexpBolometricTerm,
    "doublexp": DoublexpBolometricTerm,
}



import numpy as np
from numpy.testing import assert_allclose

from light_curve.light_curve_py import ReducedChi2


def test_redchi2_equal_sigma():
    m = np.array([1.0, 1.0, 2.0, 3.0, 4.0, 5.0])
    sigma = np.array([0.5, 0.5, 0.5, 0.5, 0.5, 0.5])
    feature = ReducedChi2()
    desired = feature(np.linspace(0, 1, len(m)), m, sigma)
    actual = 10.666667
    assert_allclose(actual, desired)


def test_redchi2_different_sigma():
    m = np.arange(6)
    sigma = np.array([0.5, 1.0, 0.5, 1.0, 0.5, 1.0])
    feature = ReducedChi2()
    desired = feature(np.linspace(0, 1, len(m)), m, sigma)
    actual = 8.48
    assert_allclose(actual, desired)



from ._base import BaseSingleBandFeature


class Skew(BaseSingleBandFeature):
    def _eval_single_band(self, t, m, sigma=None):
        try:
            from scipy.stats import skew
        except ImportError:
            raise ImportError("scipy is required for Skew feature, please install it")

        return skew(m, bias=False)

    @property
    def size_single_band(self):
        return 1


__all__ = ("Skew",)



import numpy as np

from ._base import BaseSingleBandFeature


class WeightedMean(BaseSingleBandFeature):
    def _eval_single_band(self, t, m, sigma=None):
        return np.average(m, weights=np.power(sigma, -2))

    @property
    def size_single_band(self):
        return 1


__all__ = ("WeightedMean",)



import numpy as np

from ._base import BaseSingleBandFeature


class PercentAmplitude(BaseSingleBandFeature):
    def _eval_single_band(self, t, m, sigma=None):
        median = np.median(m)
        return np.max((np.max(m) - median, median - np.min(m)))

    @property
    def size_single_band(self):
        return 1


__all__ = ("PercentAmplitude",)



import copy
import pickle

import pytest

from light_curve.light_curve_ext import ln_prior

LN_PRIORS = [
    ln_prior.none(),
    ln_prior.log_normal(1.0, 2.0),
    ln_prior.log_uniform(1.0, 10.0),
    ln_prior.normal(-1.0, 3.0),
    ln_prior.uniform(-2.0, 1.0),
    ln_prior.mix([(0.5, ln_prior.uniform(0.0, 1.0)), (0.5, ln_prior.normal(0.5, 0.1))]),
]


@pytest.mark.parametrize("lnpr", LN_PRIORS)
@pytest.mark.parametrize("pickle_protocol", tuple(range(2, pickle.HIGHEST_PROTOCOL + 1)))
def test_pickle(lnpr, pickle_protocol):
    b = pickle.dumps(lnpr, protocol=pickle_protocol)
    pickle.loads(b)


@pytest.mark.parametrize("lnpr", LN_PRIORS)
def test_copy(lnpr):
    copy.copy(lnpr)


@pytest.mark.parametrize("lnpr", LN_PRIORS)
def test_deepcopy(lnpr):
    copy.deepcopy(lnpr)



def patch_astropy_for_feets():
    """Feets is incompatible with astropy v6.0 because of backward incompatible
    changes in the subpackage structure. This function monkey patches astropy
    to make it compatible with feets.
    """
    import importlib
    import sys
    from importlib.metadata import version

    try:
        astropy_version = version("astropy")
    except ImportError:
        # astropy is not installed
        return
    if int(astropy_version.split(".")[0]) < 6:
        # astropy is older than v6.0
        return

    lombscargle = importlib.import_module("astropy.timeseries.periodograms.lombscargle")
    sys.modules["astropy.stats.lombscargle"] = lombscargle


patch_astropy_for_feets()



import numpy as np

from ._base import BaseSingleBandFeature


class ReducedChi2(BaseSingleBandFeature):
    def _eval_single_band(self, t, m, sigma=None):
        n = len(m)
        weights = sigma**-2
        m_wmean = np.average(m, weights=weights)
        s = (m - m_wmean) ** 2 * weights
        return np.sum(s) / (n - 1)

    @property
    def size_single_band(self):
        return 1


__all__ = ("ReducedChi2",)



from ._base import BaseSingleBandFeature


class AndersonDarlingNormal(BaseSingleBandFeature):
    def _eval_single_band(self, t, m, sigma=None):
        try:
            from scipy.stats import anderson
        except ImportError:
            raise ImportError("scipy is required for AndersonDarlingNormal feature, please install it")

        n = len(m)
        return anderson(m).statistic * (1 + 4 / n - 25 / n**2)

    @property
    def size_single_band(self):
        return 1


__all__ = ("AndersonDarlingNormal",)



from dataclasses import dataclass
from typing import Dict, Union

import numpy as np
from numpy.typing import NDArray

__all__ = ["Scaler", "MultiBandScaler"]


@dataclass()
class Scaler:
    """Shift and scale arrays"""

    shift: Union[float, NDArray]
    """Scale to apply to the array

    Either a single value or an array of the same shape as the input array
    """

    scale: Union[float, NDArray]
    """Scale to apply to the array

    Either a single value or an array of the same shape as the input array
    """

    @classmethod
    def from_time(cls, t) -> "Scaler":
        """Create a Scaler from a time array

        It just computes the mean and standard deviation of the array.
        """
        shift = np.mean(t)
        scale = np.std(t)
        if scale == 0.0:
            scale = 1.0
        return cls(shift=shift, scale=scale)

    def do_shift_scale(self, x):
        return (x - self.shift) / self.scale

    def undo_shift_scale(self, x):
        return x * self.scale + self.shift

    def do_scale(self, x):
        return x / self.scale

    def undo_scale(self, x):
        return x * self.scale


@dataclass()
class MultiBandScaler(Scaler):
    """Shift and scale arrays, optionally per band"""

    per_band_shift: Dict[str, float]
    """Shift to apply to each band"""

    @classmethod
    def from_flux(cls, flux, band, *, with_baseline: bool) -> "MultiBandScaler":
        """Create a Scaler from a flux array.

        It uses standard deviation for the scale. For the shift, it is either
        zero (`with_baseline=False`) or the mean of each band otherwise.
        """
        uniq_bands = np.unique(band)
        per_band_shift = dict.fromkeys(uniq_bands, 0.0)
        shift_array = np.zeros(len(flux))

        if with_baseline:
            for b in uniq_bands:
                idx = band == b
                shift_array[idx] = per_band_shift[b] = np.mean(flux[idx])

        scale = np.std(flux)
        if scale == 0.0:
            scale = 1.0

        return cls(shift=shift_array, scale=scale, per_band_shift=per_band_shift)

    def undo_shift_scale_band(self, x, band):
        return x * self.scale + self.per_band_shift.get(band, 0)



import numpy as np
from numpy.testing import assert_allclose

from light_curve.light_curve_py import StandardDeviation


def test_stdev():
    m = np.arange(10)
    feature = StandardDeviation()
    actual = feature(m, m, None)
    m_sum = sum((m - np.mean(m)) ** 2)
    desired = np.sqrt(m_sum / (len(m) - 1))
    assert_allclose(actual, desired)



import numpy as np
from numpy.testing import assert_allclose

from light_curve.light_curve_py import Mean


def test_mean_1():
    n = 10
    feature = Mean()
    m = np.arange(10)
    desired = sum(m) / n
    actual = feature(m, m, None)
    assert_allclose(actual, desired)


def test_mean_2():
    feature = Mean()
    m = np.linspace(0, 50, 100)
    desired = sum(m) / len(m)
    actual = feature(m, m, None)
    assert_allclose(actual, desired)



from dataclasses import dataclass

from ._base import BaseMultiBandFeature
from .median import Median


@dataclass()
class ColorOfMedian(BaseMultiBandFeature):
    """Difference of median magnitudes in two bands."""

    blue_band: str
    red_band: str

    @property
    def is_band_required(self) -> bool:
        return True

    @property
    def is_multiband_supported(self) -> bool:
        return True

    @property
    def size(self) -> int:
        return 1

    def __post_init__(self) -> None:
        super().__post_init__()
        self.median_feature = Median(bands=[self.blue_band, self.red_band])

    def _eval_and_fill(self, *, t, m, sigma, band, fill_value):
        median = self.median_feature._eval_and_fill(t=t, m=m, sigma=sigma, band=band, fill_value=fill_value)
        return median[0] - median[1]



import numpy as np

from ._base import BaseSingleBandFeature


class Kurtosis(BaseSingleBandFeature):
    def _eval_single_band(self, t, m, sigma=None):
        n = len(m)
        m_mean = np.mean(m)
        m_st = np.std(m, ddof=1) ** 4
        m_sum = sum(np.power(m - m_mean, 4))
        return (n * (n + 1) * m_sum) / ((n - 1) * (n - 2) * (n - 3) * m_st) - 3 * np.power((n - 1), 2) / (
            (n - 2) * (n - 3)
        )

    @property
    def size_single_band(self):
        return 1


__all__ = ("Kurtosis",)



import numpy as np

from light_curve.light_curve_py import RainbowFit


def test_noisy_with_baseline():
    rng = np.random.default_rng(0)

    band_wave_aa = {"g": 4770.0, "r": 6231.0, "i": 7625.0, "z": 9134.0}

    reference_time = 60000.0
    amplitude = 1.0
    rise_time = 5.0
    fall_time = 30.0
    Tmin = 5e3
    Tmax = 15e3
    k_sig = 4.0
    baselines = {b: 0.3 * amplitude + rng.exponential(scale=0.3 * amplitude) for b in band_wave_aa}

    expected = [reference_time, amplitude, rise_time, fall_time, Tmin, Tmax, k_sig, *baselines.values(), 1.0]

    feature = RainbowFit.from_angstrom(band_wave_aa, with_baseline=True, temperature="sigmoid", bolometric="bazin")

    t = np.sort(rng.uniform(reference_time - 3 * rise_time, reference_time + 3 * fall_time, 1000))
    band = rng.choice(list(band_wave_aa), size=len(t))

    flux = feature.model(t, band, *expected)
    # S/N = 10 for minimum flux, scale for Poisson noise
    flux_err = np.sqrt(flux * np.min(flux)) / 10.0
    flux += rng.normal(0.0, flux_err)

    actual = feature(t, flux, sigma=flux_err, band=band)

    # import matplotlib.pyplot as plt
    # plt.scatter(t, flux, s=5, label="data")
    # plt.errorbar(t, flux, yerr=flux_err, ls="none", capsize=1)
    # plt.plot(t, feature.model(t, band, *expected), "x", label="expected")
    # plt.plot(t, feature.model(t, band, *actual), "*", label="actual")
    # plt.legend()
    # plt.show()

    np.testing.assert_allclose(actual[:-1], expected[:-1], rtol=0.1)


def test_noisy_all_functions_combination():
    rng = np.random.default_rng(0)
    band_wave_aa = {"g": 4770.0, "r": 6231.0, "i": 7625.0, "z": 9134.0}
    t = np.sort(rng.uniform(59985.0, 60090.0, 1000))
    band = rng.choice(list(band_wave_aa), size=len(t))

    bazin_parameters = [
        60000.0,  # reference_time
        1.0,  # amplitude
        5.0,  # rise_time
        30.0,  # fall_time
    ]

    sigmoid_parameters = [
        60000.0,  # reference_time
        1.0,  # amplitude
        5.0,  # rise_time
    ]

    linexp_parameters = [
        60000.0,  # reference_time
        1,  # amplitude
        -20,  # rise_time
    ]

    doublexp_parameters = [60000.0, 10, 5, 10, 0.1]  # reference_time  # amplitude  # time1  # time2  # p

    bolometric_names = ["bazin", "sigmoid", "linexp", "doublexp"]
    bolometric_params = [bazin_parameters, sigmoid_parameters, linexp_parameters, doublexp_parameters]

    Tsigmoid_parameters = [5e3, 15e3, 10]  # Tmin  # Tmax  # t_color

    constant_parameters = [1e4]  # T

    temperature_names = ["constant", "sigmoid"]
    temperature_params = [constant_parameters, Tsigmoid_parameters]

    for idx_b in range(len(bolometric_names)):
        for idx_t in range(len(temperature_names)):

            expected = [*bolometric_params[idx_b], *temperature_params[idx_t], 1.0]

            feature = RainbowFit.from_angstrom(
                band_wave_aa,
                with_baseline=False,
                temperature=temperature_names[idx_t],
                bolometric=bolometric_names[idx_b],
            )

            flux = feature.model(t, band, *expected)

            # The linexp function can reach unphysical negative flux values
            protected_flux = np.where(flux > 1e-3, flux, 1e-3)

            # S/N = 10 for minimum flux, scale for Poisson noise
            flux_err = np.sqrt(protected_flux * np.min(protected_flux)) / 10.0
            flux += rng.normal(0.0, flux_err)

            actual = feature(t, flux, sigma=flux_err, band=band)

            # import matplotlib.pyplot as plt
            # plt.figure()
            # plt.scatter(t, flux, s=5, label="data")
            # plt.errorbar(t, flux, yerr=flux_err, ls="none", capsize=1)
            # plt.plot(t, feature.model(t, band, *expected), "x", label="expected")
            # plt.plot(t, feature.model(t, band, *actual), "*", label="actual")
            # plt.ylim(-.05, flux.max()+0.1)
            # plt.legend()
            # plt.show()

            np.testing.assert_allclose(actual[:-1], expected[:-1], rtol=0.1)



import numpy as np

from ._base import BaseSingleBandFeature


class Cusum(BaseSingleBandFeature):
    def _eval_single_band(self, t, m, sigma=None):
        m_mean = np.mean(m)
        m_std = np.std(m, ddof=1)
        m_new = np.cumsum(m - m_mean)
        result = m_new / (len(m) * m_std)
        return np.ptp(result)

    @property
    def size_single_band(self):
        return 1


__all__ = ("Cusum",)



from dataclasses import dataclass

import numpy as np

from ..dataclass_field import dataclass_field
from ._base import BaseSingleBandFeature


@dataclass()
class PercentDifferenceMagnitudePercentile(BaseSingleBandFeature):
    quantile: float = dataclass_field(default=0.25, kw_only=True)

    def _eval_single_band(self, t, m, sigma=None):
        try:
            from scipy.stats.mstats import mquantiles
        except ImportError:
            raise ImportError("scipy is required for PercentDifferenceMagnitudePercentile feature, please install it")

        median = np.median(m)
        q1, q2 = mquantiles(m, [self.quantile, 1 - self.quantile], alphap=0.5, betap=0.5)
        return (q2 - q1) / median

    @property
    def size_single_band(self):
        return 1


__all__ = ("PercentDifferenceMagnitudePercentile",)



import numpy as np

from ._base import BaseSingleBandFeature


class OtsuSplit(BaseSingleBandFeature):
    """Otsu threshholding algorithm

    Difference of subset means, standard deviation of the lower subset, standard deviation
    of the upper subset and lower-to-all observation count ratio for two subsets of magnitudes
    obtained by Otsu's method split.

    Otsu's method is used to perform automatic thresholding. The algorithm returns a single
    threshold that separate values into two classes. This threshold is determined by minimizing
    intra-class intensity variance, or equivalently, by maximizing inter-class variance.

    - Depends on: **magnitude**
    - Minimum number of observations: **2**
    - Number of features: **4**

    Otsu, Nobuyuki 1979. [DOI:10.1109/tsmc.1979.4310076](https://doi.org/10.1109/tsmc.1979.4310076)
    """

    def _eval_single_band(self, t, m, sigma=None):
        n = len(m)
        m = np.sort(m)
        arg, mean0, mean1 = self._threshold_arg(m)

        std_lower = np.std(m[: arg + 1], ddof=1)
        std_upper = np.std(m[arg + 1 :], ddof=1)

        if len(m[: arg + 1]) == 1:
            std_lower = 0
        if len(m[arg + 1 :]) == 1:
            std_upper = 0

        lower_to_all_ratio = (arg + 1) / n

        return mean1[arg] - mean0[arg], std_lower, std_upper, lower_to_all_ratio

    @staticmethod
    def _threshold_arg(sorted_m):
        n = len(sorted_m)
        amounts = np.arange(1, n)

        w0 = amounts / n
        w1 = 1 - w0

        cumsum0 = np.cumsum(sorted_m)[:-1]
        cumsum1 = np.cumsum(sorted_m[::-1])[:-1][::-1]
        mean0 = cumsum0 / amounts
        mean1 = cumsum1 / amounts[::-1]

        inter_class_variance = w0 * w1 * (mean0 - mean1) ** 2
        arg = np.argmax(inter_class_variance)
        return arg, mean0, mean1

    @staticmethod
    def threshold(m):
        """The Otsu threshold method."""
        m = np.sort(m)
        arg, _, _ = OtsuSplit._threshold_arg(m)
        return m[arg + 1]

    @property
    def names(self):
        return "otsu_mean_diff", "otsu_std_lower", "otsu_std_upper", "otsu_lower_to_all_ratio"

    @property
    def descriptions(self):
        return (
            "difference between mean values of Otsu split subsets",
            "standard deviation for observations below the threshold given by Otsu method",
            "standard deviation for observations above the threshold given by Otsu method",
            "ratio of quantity of observations bellow the threshold given by Otsu method to quantity of all observations",  # noqa E501
        )

    @property
    def size_single_band(self):
        return 4


__all__ = ("OtsuSplit",)



from abc import abstractmethod
from dataclasses import dataclass
from typing import Dict, List, Union

import numpy as np

__all__ = ["temperature_terms", "BaseTemperatureTerm", "ConstantTemperatureTerm", "SigmoidTemperatureTerm"]


@dataclass()
class BaseTemperatureTerm:
    """Bolometric term for the Rainbow"""

    @staticmethod
    @abstractmethod
    def parameter_names() -> List[str]:
        return NotImplementedError

    @staticmethod
    @abstractmethod
    def parameter_scalings() -> List[Union[str, None]]:
        """Describes how to unscale the parameters.

        Should be the list the same shape as returned by `parameter_names()`, and describes
        how the parameters should be un-scaled from the fit done in scaled coordinates.

        List items should be either None or one of the following strings:
        - time - the parameter is scaled and shifted like measurement times
        - timescale - the parameter is scaled like measurement times, but not shifted, thus
            behaving like a difference between two measurement times
        - flux - the parameter is scaled like the flux points, without additional shifts
            applied to them. Suitable for amplitude-like parameters.
        - None - the parameter is kept as is, without any additional scaling or shifting
        """
        return NotImplementedError

    @staticmethod
    @abstractmethod
    def value(self, t, *params) -> float:
        return NotImplementedError

    @staticmethod
    @abstractmethod
    def initial_guesses(t, m, sigma, band) -> Dict[str, float]:
        return NotImplementedError

    @staticmethod
    @abstractmethod
    def limits(t, m, sigma, band) -> Dict[str, float]:
        return NotImplementedError


@dataclass
class ConstantTemperatureTerm(BaseTemperatureTerm):
    """Constant temperature"""

    @staticmethod
    def parameter_names():
        return ["T"]

    @staticmethod
    def parameter_scalings():
        return [None]

    @staticmethod
    def value(t, temp):
        return np.full_like(t, temp)

    @staticmethod
    def initial_guesses(t, m, sigma, band):
        initial = {}
        initial["T"] = 8000.0

        return initial

    @staticmethod
    def limits(t, m, sigma, band):
        limits = {}
        limits["T"] = (1e2, 2e6)  # K

        return limits


@dataclass
class SigmoidTemperatureTerm(BaseTemperatureTerm):
    """Sigmoid temperature"""

    @staticmethod
    def parameter_names():
        return ["reference_time", "Tmin", "Tmax", "t_color"]

    @staticmethod
    def parameter_scalings():
        return ["time", None, None, "timescale"]

    @staticmethod
    def value(t, t0, temp_min, temp_max, t_color):
        dt = t - t0

        # To avoid numerical overflows, let's only compute the exponent not too far from t0
        idx1 = dt <= -100 * t_color
        idx2 = (dt > -100 * t_color) & (dt < 100 * t_color)
        idx3 = dt >= 100 * t_color

        result = np.zeros(len(dt))
        result[idx1] = temp_max
        result[idx2] = temp_min + (temp_max - temp_min) / (1.0 + np.exp(dt[idx2] / t_color))
        result[idx3] = temp_min

        return result

    @staticmethod
    def initial_guesses(t, m, sigma, band):
        med_dt = median_dt(t, band)

        initial = {}
        initial["Tmin"] = 7000.0
        initial["Tmax"] = 10000.0
        initial["t_color"] = 10 * med_dt

        return initial

    @staticmethod
    def limits(t, m, sigma, band):
        t_amplitude = np.ptp(t)
        med_dt = median_dt(t, band)

        limits = {}
        limits["Tmin"] = (1e3, 2e6)  # K
        limits["Tmax"] = (1e3, 2e6)  # K
        limits["t_color"] = (2 * med_dt, 10 * t_amplitude)

        return limits


@dataclass
class DelayedSigmoidTemperatureTerm(BaseTemperatureTerm):
    """Sigmoid temperature with delay w.r.t. bolometric peak"""

    @staticmethod
    def parameter_names():
        return ["reference_time", "Tmin", "Tmax", "t_color", "t_delay"]

    @staticmethod
    def parameter_scalings():
        return ["time", None, None, "timescale", "timescale"]

    @staticmethod
    def value(t, t0, Tmin, Tmax, t_color, t_delay):
        dt = t - t0 - t_delay

        # To avoid numerical overflows, let's only compute the exponent not too far from t0
        idx1 = dt <= -100 * t_color
        idx2 = (dt > -100 * t_color) & (dt < 100 * t_color)
        idx3 = dt >= 100 * t_color

        result = np.zeros(len(dt))
        result[idx1] = Tmax
        result[idx2] = Tmin + (Tmax - Tmin) / (1.0 + np.exp(dt[idx2] / t_color))
        result[idx3] = Tmin

        return result

    @staticmethod
    def initial_guesses(t, m, sigma, band):
        med_dt = median_dt(t, band)

        initial = {}
        initial["Tmin"] = 7000.0
        initial["Tmax"] = 10000.0
        initial["t_color"] = 10 * med_dt
        initial["t_delay"] = 0.0

        return initial

    @staticmethod
    def limits(t, m, sigma, band):
        t_amplitude = np.ptp(t)
        med_dt = median_dt(t, band)

        limits = {}
        limits["Tmin"] = (1e3, 2e6)  # K
        limits["Tmax"] = (1e3, 2e6)  # K
        limits["t_color"] = (2 * med_dt, 10 * t_amplitude)
        limits["t_delay"] = (-t_amplitude, t_amplitude)

        return limits


def median_dt(t, band):
    # Compute the median distance between points in each band
    dt = []
    for b in np.unique(band):
        dt += list(t[band == b][1:] - t[band == b][:-1])
    med_dt = np.median(dt)
    return med_dt


temperature_terms = {
    "constant": ConstantTemperatureTerm,
    "sigmoid": SigmoidTemperatureTerm,
    "delayed_sigmoid": DelayedSigmoidTemperatureTerm,
}



import numpy as np

from ._base import BaseSingleBandFeature


class Roms(BaseSingleBandFeature):
    r"""Robust median statistic

    $$
    \frac1{N-1} \sum_{i=0}^{N-1} \frac{|m_i - \mathrm{median}(m_i)|}{\sigma_i}
    $$
    For non-variable data, it should be less than one.

    - Depends on: **magnitude**, **errors**
    - Minimum number of observations: **2**
    - Number of features: **1**

    Enoch, Brown, Burgasser 2003. [DOI:10.1086/376598](https://www.doi.org/10.1086/376598)
    """

    def _eval_single_band(self, t, m, sigma=None):
        n = len(m)
        median = np.median(m)
        return np.sum(np.abs(m - median) / sigma) / (n - 1)

    @property
    def size_single_band(self):
        return 1


__all__ = ("Roms",)



import numpy as np
from numpy.testing import assert_allclose

from light_curve.light_curve_py import Eta, EtaE


def test_etae_1():
    a = np.array([1, 2, 3, 8])
    t = np.array([1, 3, 5, 7])
    feature = EtaE()
    feature_2 = Eta()
    actual = feature(t, a, None)
    actual_2 = feature_2(t, a, None)
    assert_allclose(actual, actual_2)


def test_etae_2():
    a = np.array([1, 2, 3, 8, 11])
    t = np.array([1, 2, 7, 9, 10])
    feature = EtaE()
    desired = 1.114434
    actual = feature(t, a, None)
    assert_allclose(actual, desired, rtol=1e-06)



import sys

if sys.version_info >= (3, 10):
    from dataclasses import field as dataclass_field
else:
    from dataclasses import field as _field

    def dataclass_field(*, kw_only, **kwargs):
        return _field(**kwargs)


__all__ = ["dataclass_field"]



import copy
import inspect
import pickle

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_array_equal

import light_curve.light_curve_ext as lc


def _feature_classes(module, *, exclude_parametric=True):
    for name, member in inspect.getmembers(module):
        if name.startswith("_"):
            continue
        if inspect.ismodule(member):
            yield from _feature_classes(member)
        if not inspect.isclass(member):
            continue
        if not issubclass(member, lc._FeatureEvaluator):
            continue
        if member is lc.JSONDeserializedFeature:
            continue
        # Skip classes with non-trivial constructors
        if exclude_parametric:
            try:
                member()
            except TypeError:
                continue
        yield member


non_param_feature_classes = frozenset(_feature_classes(lc, exclude_parametric=True))
assert len(non_param_feature_classes) > 0

all_feature_classes = frozenset(_feature_classes(lc, exclude_parametric=False))
assert len(all_feature_classes) > 0


def get_new_args_kwargs(cls):
    if hasattr(cls, "__getnewargs_ex__"):
        return cls.__getnewargs_ex__()
    if hasattr(cls, "__getnewargs__"):
        args = cls.__getnewargs__()
        return args, {}
    return (), {}


def new_default(cls, **kwargs):
    args, kwargs_ = get_new_args_kwargs(cls)
    kwargs = dict(kwargs_, **kwargs)
    return cls(*args, **kwargs)


def construct_example_objects(cls, *, parametric_variants=1, rng=None):
    # Extractor is special
    if cls is lc.Extractor:
        return [cls(lc.BeyondNStd(1.5), lc.LinearFit())]

    # No mandatory arguments
    if not hasattr(cls, "__getnewargs__"):
        return [cls()]

    # default mandatory arguments
    args, kwargs = get_new_args_kwargs(cls)

    # Add Mean feature for metafeatures
    args = [[lc.Mean()] if arg == () else arg for arg in args]

    objects = [cls(*args, **kwargs)]
    # Nothing to mutate
    if not any(isinstance(arg, float) for arg in args + list(kwargs.values())):
        return objects

    # Mutate floats
    rng = np.random.default_rng(rng)

    def mutation(value):
        if not isinstance(value, float):
            return value
        return value * rng.uniform(0.9, 1.1) + rng.uniform(0.0, 1e-3)

    for _ in range(1, parametric_variants):
        mutated_args = list(map(mutation, args))
        mutated_kwargs = {name: mutation(value) for name, value in kwargs.items()}
        objects.append(cls(*mutated_args, **mutated_kwargs))
    return objects


def gen_feature_evaluators(*, parametric_variants=0, rng=None):
    if parametric_variants == 0:
        for cls in non_param_feature_classes:
            yield cls()
        return
    rng = np.random.default_rng(rng)
    for cls in all_feature_classes:
        yield from construct_example_objects(cls, parametric_variants=parametric_variants, rng=rng)


def gen_lc(n, rng=None):
    rng = np.random.default_rng(rng)

    t = np.sort(rng.normal(0, 1, n))
    m = t.copy()
    sigma = np.full_like(t, 0.1)

    return t, m, sigma


@pytest.mark.parametrize("cls", list(all_feature_classes))
def test_available_transforms(cls):
    # All available features should consume transform=None
    none = new_default(cls, transform=None)

    # If transform consumes False it
    # 1) should give the same feature as transform=None
    # 2) should be able to consume transform=True
    try:
        false = new_default(cls, transform=False)
    except NotImplementedError:
        return
    # It would be better to compare objects themselves, but __eq__ is not implemented yet
    # https://github.com/light-curve/light-curve-python/issues/148
    assert false.names == none.names
    true = new_default(cls, transform=True)
    # Check if transform=True is not the same as transform=False
    default_transform = getattr(cls, "default_transform", None)
    if default_transform != "identity":
        assert true.names != false.names

    # Both attributes should be present or absent
    assert hasattr(cls, "supported_transforms") == hasattr(cls, "default_transform")

    if not hasattr(cls, "supported_transforms"):
        return

    assert cls.default_transform in cls.supported_transforms

    for transform in cls.supported_transforms + ["default"]:
        new_default(cls, transform=transform)


@pytest.mark.parametrize("feature", gen_feature_evaluators(parametric_variants=2))
def test_negative_strides(feature):
    t = np.linspace(1, 0, 20)[::-2]
    m = np.exp(t)[:]
    err = np.random.uniform(0.1, 0.2, t.shape)
    feature(t, m, err)


# We don't want *Fit features here: not precise
@pytest.mark.parametrize("feature", gen_feature_evaluators(parametric_variants=0))
def test_float32_vs_float64(feature):
    rng = np.random.default_rng(0)
    n = 128

    t, m, sigma = gen_lc(n, rng=rng)

    results = [
        feature(t.astype(dtype), m.astype(dtype), sigma.astype(dtype), sorted=True)
        for dtype in [np.float32, np.float64]
    ]
    assert_allclose(*results, rtol=1e-5, atol=1e-5)


# We don't want *Fit features here: too slow
@pytest.mark.parametrize("feature", gen_feature_evaluators(parametric_variants=0))
def test_many_vs_call(feature):
    rng = np.random.default_rng(0)
    n_obs = 128
    n_lc = 128

    lcs = [gen_lc(n_obs, rng=rng) for _ in range(n_lc)]

    call = np.stack([feature(*lc, sorted=True) for lc in lcs])
    many = feature.many(lcs, sorted=True, n_jobs=2)
    assert_array_equal(many, call)


def test_fill_value_not_enough_observations():
    n = 1
    t = np.linspace(0.0, 1.0, n)
    m = t.copy()
    fill_value = -100.0
    sigma = np.ones_like(t)
    feature = lc.Kurtosis()
    with pytest.raises(ValueError):
        feature(t, m, sigma, fill_value=None)
    assert_array_equal(feature(t, m, sigma, fill_value=fill_value), fill_value)


@pytest.mark.parametrize("cls", all_feature_classes)
def test_nonempty_docstring(cls):
    assert len(cls.__doc__) > 10


@pytest.mark.parametrize("feature", gen_feature_evaluators(parametric_variants=2))
def test_check_t(feature):
    n_obs = 128
    t, m, sigma = gen_lc(n_obs)
    t[0] = np.nan
    with pytest.raises(ValueError):
        feature(t, m, sigma, check=True)
    t[0] = np.inf
    with pytest.raises(ValueError):
        feature(t, m, sigma, check=True)


@pytest.mark.parametrize("feature", gen_feature_evaluators(parametric_variants=2))
def test_check_m(feature):
    n_obs = 128
    t, m, sigma = gen_lc(n_obs)
    m[0] = np.nan
    with pytest.raises(ValueError):
        feature(t, m, sigma, check=True)
    m[0] = np.inf
    with pytest.raises(ValueError):
        feature(t, m, sigma, check=True)


# We need evaluators which use sigma
@pytest.mark.parametrize("cls", (lc.ExcessVariance, lc.LinearFit, lc.ReducedChi2, lc.StetsonK, lc.WeightedMean))
def test_check_sigma(cls):
    n_obs = 128
    t, m, sigma = gen_lc(n_obs)
    sigma[0] = np.nan
    feature = cls()
    with pytest.raises(ValueError):
        feature(t, m, sigma, check=True)
    # infinite values are allowed for sigma
    sigma[0] = np.inf
    feature(t, m, sigma, check=True)


@pytest.mark.parametrize("feature", gen_feature_evaluators(parametric_variants=5, rng=None))
@pytest.mark.parametrize("pickle_protocol", tuple(range(2, pickle.HIGHEST_PROTOCOL + 1)))
def test_pickling(feature, pickle_protocol):
    n_obs = 128
    data = gen_lc(n_obs)
    values = feature(*data)

    b = pickle.dumps(feature, protocol=pickle_protocol)
    new_feature = pickle.loads(b)

    new_values = new_feature(*data)
    assert_array_equal(values, new_values)


@pytest.mark.parametrize("feature", gen_feature_evaluators(parametric_variants=5, rng=None))
def test_copy_deepcopy(feature):
    n_obs = 128
    data = gen_lc(n_obs)
    values = feature(*data)

    copied = copy.copy(feature)
    values_copied = copied(*data)
    assert_array_equal(values, values_copied)

    deepcopied = copy.deepcopy(feature)
    values_deepcopied = deepcopied(*data)
    assert_array_equal(values, values_deepcopied)


PICKLE_BENCHMARK_FEATURES = [
    lc.Amplitude(),  # no parameters
    lc.BeyondNStd(1.5),  # parametric
    lc.Extractor(  # large
        lc.Amplitude(),
        lc.BeyondNStd(2.0),
        lc.Bins(
            [lc.Kurtosis(), lc.LinearTrend(), lc.WeightedMean()],
            window=2.0,
            offset=59500.5,
        ),
        lc.Periodogram(features=[lc.InterPercentileRange(0.01)], peaks=5, max_freq_factor=12.0),
    ),
]


@pytest.mark.parametrize("feature", PICKLE_BENCHMARK_FEATURES)
def test_benchmark_pickle_loads(feature, benchmark):
    b = pickle.dumps(feature, protocol=pickle.HIGHEST_PROTOCOL)
    benchmark(pickle.loads, b)


@pytest.mark.parametrize("feature", PICKLE_BENCHMARK_FEATURES)
def test_benchmark_pickle_dumps(feature, benchmark):
    benchmark(pickle.dumps, feature, protocol=pickle.HIGHEST_PROTOCOL)


@pytest.mark.parametrize("feature", PICKLE_BENCHMARK_FEATURES)
def test_benchmark_copy(feature, benchmark):
    benchmark(copy.copy, feature)


@pytest.mark.parametrize("feature", PICKLE_BENCHMARK_FEATURES)
def test_benchmark_deepcopy(feature, benchmark):
    benchmark(copy.deepcopy, feature)


# We do not check pure MCMC because it requires a lot of iterations and would be too slow
@pytest.mark.parametrize("algo", ("ceres", "mcmc-ceres", "lmsder", "mcmc-lmsder"))
def test_bazin_fit_precise(algo):
    bazin = lc.BazinFit(algo)

    true_params = np.array([10.0, -2.0, 10.0, 10.0, 25.0])
    t = np.linspace(-50.0, 120.0, 1000)
    flux = bazin.model(t, true_params)
    fluxerr = np.ones_like(t)

    *params, reduced_chi2 = bazin(t, flux, fluxerr)
    assert_allclose(true_params, params, rtol=1e-4)  # tolerance set to underlying algorithms


@pytest.mark.parametrize("feature", gen_feature_evaluators(parametric_variants=5, rng=None))
def test_json_serialization(feature):
    n_obs = 128
    data = gen_lc(n_obs)
    values = feature(*data)

    from_to_json = lc.feature_from_json(feature.to_json())
    values_from_to_json = from_to_json(*data)
    assert_array_equal(values, values_from_to_json)


def test_json_deserialization():
    json = """
    {"FeatureExtractor":{"features":[{"Transformed":{"feature":{"AndersonDarlingNormal":{}},"transformer":{"Ln1p":{}}}},
    {"Transformed":{"feature":{"BazinFit":{"algorithm":{"Ceres":{"loss_factor":null,"niterations":20}},"inits_bounds":
    {"OptionArrays":{"init":[null,null,null,null,null],"lower":[0.0036307805477010066,null,null,0.0001,0.0001],"upper":
    [3630780547.7010174,null,null,30000.0,30000.0]}},"ln_prior":{"Fixed":{"None":{}}}}},"transformer":{"BazinFit":
    {"mag_zp":23.899999618530273}}}},{"ExcessVariance":{}}]}}
    """
    from_json = lc.feature_from_json(json)
    assert isinstance(from_json, lc._FeatureEvaluator)
    from_json(*gen_lc(128))



from typing import Any, Dict

import numpy as np
from numpy.typing import ArrayLike, NDArray


class Bands:
    """Monochromatic passbands for Rainbow fit."""

    @property
    def names(self) -> NDArray:
        """Names of the bands"""
        return self._names.copy()

    @property
    def index(self) -> NDArray[int]:
        """Index array for the bands"""
        return np.arange(len(self._names))

    def wave_cm(self, band) -> float:
        """Wavelength of the band in cm"""
        return self._names_to_wave_cm[band]

    @property
    def mean_wave_cm(self) -> float:
        """Mean wavelength of the bands in cm"""
        return np.mean(self._wave_cm)

    def index_to_wave_cm(self, index: NDArray[int]) -> NDArray[float]:
        """Wavelength of the band in cm"""
        return self._wave_cm[index]

    def __init__(self, names: ArrayLike, wave_cm: ArrayLike):
        self._input_validation(names, wave_cm)

        self._names = np.asarray(names)
        self._wave_cm = np.asarray(wave_cm)

        self._names_to_wave_cm = dict(zip(self._names, self._wave_cm))

        self._name_to_index = dict(zip(self._names, range(len(self._names))))
        self.get_index = np.vectorize(self._name_to_index.get)

    @classmethod
    def from_dict(cls, band_wave_cm: Dict[Any, float]) -> "Bands":
        """Create Bands from a dictionary"""
        names, wave_cm = zip(*band_wave_cm.items())
        return cls(names, wave_cm)

    @staticmethod
    def _input_validation(names: ArrayLike, wave_cm: ArrayLike):
        if len(names) != len(wave_cm):
            raise ValueError("names and wave_cm must have the same length")

        if len(names) == 0:
            raise ValueError("At least one band must be specified.")

        if len(set(names)) != len(names):
            raise ValueError("names must be unique")

        if any(lmbd <= 0 for lmbd in wave_cm):
            raise ValueError("wave_cm must be positive")



"""Maximum-likelihood based cost function"""

from typing import Callable, Dict, Tuple

import numpy as np

try:
    from iminuit import Minuit
    from scipy.special import erf
except ImportError:
    MaximumLikelihood = None
else:

    class MaximumLikelihood:
        errordef = Minuit.LIKELIHOOD

        def __init__(
            self, model: Callable, parameters: Dict[str, Tuple[float, float]], upper_mask=None, *, x, y, yerror
        ):
            self.model = model
            self.x = x
            self.y = y
            self.yerror = yerror
            self.upper_mask = upper_mask

            self._parameters = parameters
            # FIXME: here we assume the order of dict keys is the same as in parameters array
            self.limits = np.array([_ for _ in parameters.values()])
            self.limits0 = self.limits[:, 0]
            self.limits1 = self.limits[:, 1]
            self.limits_scale = self.limits[:, 1] - self.limits[:, 0]

        @property
        def ndata(self):
            return len(self.y)

        def __call__(self, *par):
            ym = self.model(self.x, *par)

            if self.upper_mask is None:
                result = -np.sum(self.logpdf(self.y, ym, self.yerror))
            else:
                # Measurements
                result = -np.sum(
                    self.logpdf(self.y[~self.upper_mask], ym[~self.upper_mask], self.yerror[~self.upper_mask])
                )
                # Upper limits, Tobit model
                # https://stats.stackexchange.com/questions/49443/how-to-model-this-odd-shaped-distribution-almost-a-reverse-j
                result += -np.sum(
                    self.logcdf((self.y[self.upper_mask] - ym[self.upper_mask]) / self.yerror[self.upper_mask])
                )

            # Barriers around parameter ranges
            # Scale is selected so that for the most of the range it is much smaller
            # than 0.5 which corresponds to 1-sigma errors
            result += 0.0001 * np.sum(self.barrier((par - self.limits0) / self.limits_scale))
            result += 0.0001 * np.sum(self.barrier((self.limits1 - par) / self.limits_scale))

            return result

        @staticmethod
        def logpdf(x, mu, sigma):
            # We do not need the second term as it does not depend on parameters
            return -(((x - mu) / sigma) ** 2) / 2  # - np.log(np.sqrt(2*np.pi) * sigma)

        @staticmethod
        def barrier(x):
            res = np.where(x > 0, 1 / x, np.inf)  # FIXME: naive barrier function

            return res

        @staticmethod
        def logcdf(x):
            # TODO: faster (maybe not so accurate, as we do not need it) implementation
            # return norm.logcdf(x)

            result = np.zeros(len(x))

            idx = x < -5
            result[idx] = -x[idx] ** 2 / 2 - 1 / x[idx] ** 2 - 0.9189385336 - np.log(-x[idx])
            result[~idx] = np.log(0.5) + np.log1p(erf(x[~idx] / np.sqrt(2)))

            return result


__all__ = ["MaximumLikelihood"]



from dataclasses import dataclass
from typing import Collection, Union

import numpy as np

from light_curve.light_curve_ext import Extractor as _RustExtractor
from light_curve.light_curve_ext import _FeatureEvaluator as _RustBaseFeature

from ..dataclass_field import dataclass_field
from ._base import BaseSingleBandFeature


@dataclass()
class _PyExtractor(BaseSingleBandFeature):
    features: Collection[Union[BaseSingleBandFeature, _RustBaseFeature]] = dataclass_field(
        default_factory=list, kw_only=True
    )

    def _eval_single_band(self, t, m, sigma=None):
        raise NotImplementedError("_eval_single_band is missed for _PyExtractor")

    def _eval_and_fill_single_band(self, *, t, m, sigma, fill_value):
        return np.concatenate([np.atleast_1d(feature(t, m, sigma, fill_value=fill_value)) for feature in self.features])

    @property
    def size_single_band(self):
        return sum(
            feature.size if isinstance(feature, _RustBaseFeature) else feature.size_single_band
            for feature in self.features
        )


class Extractor:
    def __new__(cls, *args: Collection[Union[BaseSingleBandFeature, _RustBaseFeature]]):
        if len(args) > 0 and all(isinstance(feature, _RustBaseFeature) for feature in args):
            return _RustExtractor(*args)
        else:
            return _PyExtractor(features=args)


__all__ = ("Extractor",)



import dataclasses
from functools import lru_cache, wraps
from itertools import count
from pathlib import Path
from typing import Generator, Iterator, List, Optional, Union

import feets
import numpy as np
import pandas as pd
import pytest
from joblib import Parallel, delayed
from numpy.testing import assert_allclose
from scipy import stats
from scipy.optimize import curve_fit

import light_curve.light_curve_ext as lc_ext
import light_curve.light_curve_py as lc_py


@dataclasses.dataclass
class Data:
    name: str
    phot_type: str
    t: np.ndarray = dataclasses.field(repr=False)
    m: np.ndarray = dataclasses.field(repr=False)
    sigma: np.ndarray = dataclasses.field(repr=False)

    # Kinda static attribute
    phot_type_choices: frozenset = dataclasses.field(
        default=frozenset(["flux", "mag"]),
        init=False,
        repr=False,
    )

    # For '*' operator
    def __iter__(self):
        def gen():
            yield self.t
            yield self.m
            yield self.sigma

        return gen()

    def __post_init__(self):
        self.name = str(self.name)
        assert self.phot_type in self.phot_type_choices
        assert self.t.size == self.m.size == self.sigma.size
        assert self.t.dtype == self.m.dtype == self.sigma.dtype


def gen_data_from_test_data_path(
    paths: Iterator[Union[Path, str]], *, n: Optional[int] = None, convert_to_flux: bool = False
) -> Generator[Data, None, None]:
    if n is None:
        take_n = count()
    else:
        take_n = range(n)
    for _, csv_file in zip(take_n, paths):
        df = pd.read_csv(csv_file)

        # Drop repeated values
        idx = np.diff(df["time"], prepend=0) > 0
        df = df[idx]

        t = df["time"].to_numpy()
        if "mag" in df.columns:
            m = df["mag"].to_numpy()
            sigma = df["magerr"].to_numpy()
            if convert_to_flux:
                m = 10 ** (-0.4 * m)
                sigma = 0.4 * np.log(10.0) * sigma * m
                phot_type = "flux"
            else:
                phot_type = "mag"
        elif "flux" in df.columns:
            m = df["flux"].to_numpy()
            sigma = df["fluxerr"].to_numpy()
            phot_type = "flux"
        else:
            raise ValueError(f"No mag neither flux column in {csv_file}")
        yield Data(csv_file, phot_type, t, m, sigma)


@lru_cache(maxsize=1)
def get_test_data_from_issues() -> List[Data]:
    data_root = Path(__file__).parent / "light-curve-test-data/from-issues"
    return list(gen_data_from_test_data_path(data_root.glob("**/*.csv")))


@lru_cache()
def get_first_n_snia_data(*, n: Optional[int] = None, convert_to_flux: bool = False) -> List[Data]:
    data_root = Path(__file__).parent / "light-curve-test-data/SNIa"
    with open(data_root / "snIa_bandg_minobs10_beforepeak3_afterpeak4.csv") as fh:
        file_names = frozenset(fh.read().split())
    light_curve_dir = data_root / "light-curves"
    # Sort to make sure the order is always the same
    csv_paths = sorted(light_curve_dir.glob("*.csv"))
    paths = (path for path in csv_paths if path.stem in file_names)
    return list(gen_data_from_test_data_path(paths, n=n, convert_to_flux=convert_to_flux))


class _Test:
    # Feature name must be updated in child classes
    name = None
    # Argument tuple for the feature constructor
    args = ()
    # And keyword arguments
    kwargs = {}

    py_feature = None

    # Specify method for a naive implementation
    naive = None

    # Specify `feets` feature name
    feets_feature = None
    feets_extractor = None
    # Specify a `str` with a reason why skip this test
    feets_skip_test = False

    # Which types of pre-generated light-curves to use
    phot_types = Data.phot_type_choices

    def setup_method(self):
        self.rust = getattr(lc_ext, self.name)(*self.args, **self.kwargs)

        try:
            self.py_feature_cls = getattr(lc_py, self.name)
        except AttributeError:
            pass
        else:
            if len(self.args) > 0:
                raise ValueError("Py features must be called with keyword arguments only")
            self.py_feature = self.py_feature_cls(**self.kwargs)

        if self.feets_feature is not None:
            self.feets_extractor = feets.FeatureSpace(only=[self.feets_feature], data=["time", "magnitude", "error"])

    # Default values of `assert_allclose`
    rtol = np.finfo(np.float32).resolution  # 1e-6
    atol = 0

    # Default values for random light curve generation
    n_obs = 1000
    t_min = 0.0
    t_max = 1000.0
    m_min = 15.0
    m_max = 21.0
    sigma_min = 0.01
    sigma_max = 0.2

    add_to_all_features = True

    def random_data(self):
        t = np.sort(np.random.uniform(self.t_min, self.t_max, self.n_obs))
        m = np.random.uniform(self.m_min, self.m_max, self.n_obs)
        sigma = np.random.uniform(self.sigma_min, self.sigma_max, self.n_obs)
        return Data("random", "mag", t, m, sigma)

    def from_issues_data(self):
        for data in get_test_data_from_issues():
            if data.phot_type not in self.phot_types:
                continue
            yield data

    def snia_data(self, n: Optional[int] = None):
        for data in get_first_n_snia_data(n=n):
            if data.phot_type not in self.phot_types:
                continue
            yield data

    def data_gen(self) -> Generator[Data, None, None]:
        yield self.random_data()
        yield from self.from_issues_data()
        yield from self.snia_data(n=10)

    def test_feature_length(self, subtests):
        for data in self.data_gen():
            with subtests.test(lc_name=data):
                result = self.rust(*data, sorted=None)
                assert len(result) == len(self.rust.names) == len(self.rust.descriptions)

    def test_close_to_lc_py(self, subtests):
        if self.py_feature is None:
            pytest.skip("No matched light_curve_py class for the feature")
        for data in self.data_gen():
            with subtests.test(data=data):
                assert_allclose(self.rust(*data), self.py_feature(*data), rtol=self.rtol, atol=self.atol)

    def test_benchmark_rust(self, benchmark):
        t, m, sigma = self.random_data()
        benchmark.group = str(type(self).__name__)
        benchmark(self.rust, t, m, sigma, sorted=True, check=False)

    @pytest.mark.nobs
    @pytest.mark.parametrize("n_obs", np.logspace(1, 6, 6))
    def test_benchmark_rust_n(self, benchmark, n_obs):
        self.n_obs = int(n_obs)
        t, m, sigma = self.random_data()

        benchmark.group = f"{n_obs}_observations"
        benchmark.name = f"{str(type(self).__name__)}_rust"
        benchmark(self.rust, t, m, sigma, sorted=True, check=False)

    def test_benchmark_lc_py(self, benchmark):
        if self.py_feature is None:
            pytest.skip("No matched light_curve_py class for the feature")

        t, m, sigma = self.random_data()

        benchmark.group = str(type(self).__name__)
        benchmark(self.py_feature, t, m, sigma, sorted=True, check=False)

    @pytest.mark.nobs
    @pytest.mark.parametrize("n_obs", np.logspace(1, 6, 6))
    def test_benchmark_lc_py_n(self, benchmark, n_obs):
        if self.py_feature is None:
            pytest.skip("No matched light_curve_py class for the feature")

        self.n_obs = int(n_obs)
        t, m, sigma = self.random_data()

        benchmark.group = f"{n_obs}_observations"
        benchmark.name = f"{str(type(self).__name__)}_py"
        benchmark(self.py_feature, t, m, sigma, sorted=True, check=False)

    def test_close_to_naive(self, subtests):
        if self.naive is None:
            pytest.skip("No naive implementation for the feature")
        for data in self.data_gen():
            with subtests.test(data=data):
                assert_allclose(self.rust(*data), self.naive(*data), rtol=self.rtol, atol=self.atol)

    def test_benchmark_naive(self, benchmark):
        if self.naive is None:
            pytest.skip("No naive implementation for the feature")

        t, m, sigma = self.random_data()

        benchmark.group = type(self).__name__
        benchmark(self.naive, t, m, sigma)

    @pytest.mark.nobs
    @pytest.mark.parametrize("n_obs", np.logspace(1, 6, 6))
    def test_benchmark_naive_n(self, benchmark, n_obs):
        if self.naive is None:
            pytest.skip("No naive implementation for the feature")

        self.n_obs = int(n_obs)
        t, m, sigma = self.random_data()

        benchmark.group = f"{n_obs}_observations"
        benchmark.name = f"{str(type(self).__name__)}_naive"
        benchmark(self.naive, t, m, sigma)

    def feets(self, t, m, sigma):
        _, result = self.feets_extractor.extract(t, m, sigma)
        return result

    def test_close_to_feets(self, subtests):
        if self.feets_extractor is None:
            pytest.skip("No feets feature provided")
        if self.feets_skip_test:
            pytest.skip("feets is expected to be different from light_curve, reason: " + self.feets_skip_test)

        for data in self.data_gen():
            with subtests.test(data=data):
                assert_allclose(self.rust(*data)[:1], self.feets(*data)[:1], rtol=self.rtol, atol=self.atol)

    def test_benchmark_feets(self, benchmark):
        if self.feets_extractor is None:
            pytest.skip("No feets feature provided")

        t, m, sigma = self.random_data()

        benchmark.group = type(self).__name__
        benchmark(self.feets, t, m, sigma)


class TestAmplitude(_Test):
    name = "Amplitude"

    def naive(self, t, m, sigma):
        return 0.5 * (np.max(m) - np.min(m))


class TestAndersonDarlingNormal(_Test):
    name = "AndersonDarlingNormal"

    feets_feature = "AndersonDarling"
    feets_skip_test = "feets uses biased statistics"

    def naive(self, t, m, sigma):
        return stats.anderson(m).statistic * (1.0 + 4.0 / m.size - 25.0 / m.size**2)


class _TestBazinFit:
    name = "BazinFit"
    kwargs = {"mcmc_niter": 1 << 10, "ceres_niter": 20, "lmsder_niter": 20}
    rtol = 1e-3  # 10x of the precision used in the feature implementation

    add_to_all_features = False  # in All* random data is used

    phot_types = frozenset(["flux"])

    @staticmethod
    def _model(t, a, b, t0, rise, fall):
        dt = t - t0
        return b + a * np.exp(-dt / fall) / (1.0 + np.exp(-dt / rise))

    def _params(self):
        a = 100
        c = 100
        t0 = 0.5 * (self.t_min + self.t_max)
        rise = 0.1 * (self.t_max - self.t_min)
        fall = 0.2 * (self.t_max - self.t_min)
        return a, c, t0, rise, fall

    # Random data yields to random results because target function has a lot of local minima
    # BTW, this test shouldn't use fixed random seed because the curve has good enough S/N to be fitted for any give
    # noise sample
    def random_data(self):
        rng = np.random.default_rng(0)
        t = np.linspace(self.t_min, self.t_max, self.n_obs)
        model = self._model(t, *self._params())
        sigma = np.sqrt(model)
        m = model + sigma * rng.normal(size=self.n_obs)
        return Data("Bazin+noise", "flux", t, m, sigma)

    # Keep random data only
    def data_gen(self) -> Generator[Data, None, None]:
        yield self.random_data()
        # All these fail because curve_fit is not good enough or because of the multiple local minima
        # yield from get_first_n_snia_data(n=10, convert_to_flux=True)

    def naive(self, t, m, sigma):
        params, _cov = curve_fit(
            self._model,
            xdata=t,
            ydata=m,
            sigma=sigma,
            xtol=self.rtol,
            # We give really good parameters' estimation!
            # p0=self._params(),
            # The same parameter estimations we use in Rust
            p0=(0.5 * np.ptp(m), np.min(m), t[np.argmax(m)], 0.5 * np.ptp(t), 0.5 * np.ptp(t)),
        )
        reduced_chi2 = np.sum(np.square((self._model(t, *params) - m) / sigma)) / (t.size - params.size)
        return_value = tuple(params) + (reduced_chi2,)
        return return_value


@pytest.mark.skip("Runs too long")
class TestBazinFitMcmc(_TestBazinFit, _Test):
    args = ("mcmc",)


if lc_ext._built_with_ceres:

    class TestBazinFitCeres(_TestBazinFit, _Test):
        args = ("ceres",)

    @pytest.mark.skip("Runs too long")
    class TestBazinFitMcmcCeres(_TestBazinFit, _Test):
        args = ("mcmc-ceres",)


if lc_ext._built_with_gsl:

    class TestBazinFitLmsder(_TestBazinFit, _Test):
        args = ("lmsder",)

    @pytest.mark.skip("Runs too long")
    class TestBazinFitMcmcLmsder(_TestBazinFit, _Test):
        args = ("mcmc-lmsder",)


class TestBeyond1Std(_Test):
    nstd = 1.0

    name = "BeyondNStd"
    kwargs = dict(nstd=nstd)

    feets_feature = "Beyond1Std"
    feets_skip_test = "feets uses biased statistics"

    def naive(self, t, m, sigma):
        mean = np.mean(m)
        interval = self.nstd * np.std(m, ddof=1)
        return np.count_nonzero(np.abs(m - mean) > interval) / m.size


class TestCusum(_Test):
    name = "Cusum"

    feets_feature = "Rcs"
    feets_skip_test = "feets uses biased statistics"


class TestEta(_Test):
    name = "Eta"

    def naive(self, t, m, sigma):
        return np.sum(np.square(m[1:] - m[:-1])) / (np.var(m, ddof=0) * m.size)


class TestEtaE(_Test):
    name = "EtaE"

    feets_feature = "Eta_e"
    feets_skip_test = "feets fixed EtaE from the original paper in different way"

    def naive(self, t, m, sigma):
        return (
            np.sum(np.square((m[1:] - m[:-1]) / (t[1:] - t[:-1])))
            * (t[-1] - t[0]) ** 2
            / (np.var(m, ddof=0) * m.size * (m.size - 1) ** 2)
        )


class TestExcessVariance(_Test):
    name = "ExcessVariance"

    def naive(self, t, m, sigma):
        return (np.var(m, ddof=1) - np.mean(sigma**2)) / np.mean(m) ** 2


class TestInterPercentileRange(_Test):
    quantile = 0.25

    name = "InterPercentileRange"
    kwargs = dict(quantile=quantile)

    feets_feature = "Q31"
    feets_skip_test = "feets uses different quantile type"


class TestOtsuSplit(_Test):
    name = "OtsuSplit"


def magnitude_function(func, *args, **kwargs):
    @wraps(func)
    def decorated(t, m, sigma=None, sorted=None, check=None):
        return func(m)

    return decorated


class TestOtsuSplitThreshold(_Test):
    def setup_method(self):
        self.py_feature = magnitude_function(lc_py.OtsuSplit.threshold)
        self.rust = magnitude_function(lc_ext.OtsuSplit.threshold)

    def test_feature_length(self):
        """Not a real feature extractor, no need to test length"""
        pass


class TestKurtosis(_Test):
    name = "Kurtosis"

    feets_feature = "SmallKurtosis"
    feets_skip_test = "feets uses equation for unbiased kurtosis, but put biased standard deviation there"

    def naive(self, t, m, sigma):
        return stats.kurtosis(m, fisher=True, bias=False)


class TestLinearTrend(_Test):
    name = "LinearTrend"

    feets_feature = "LinearTrend"

    def naive(self, t, m, sigma):
        (slope, _), ((slope_sigma2, _), _) = np.polyfit(t, m, deg=1, cov=True)
        sigma_noise = np.sqrt(np.polyfit(t, m, deg=1, full=True)[1][0] / (t.size - 2))
        return np.array([slope, np.sqrt(slope_sigma2), sigma_noise])


def generate_test_magnitude_percentile_ratio(quantile_numerator, quantile_denominator, feets_feature):
    return type(
        f"TestMagnitudePercentageRatio{int(quantile_numerator * 100):d}",
        (_Test,),
        dict(
            kwargs=dict(quantile_numerator=quantile_numerator, quantile_denominator=quantile_denominator),
            quantile_numerator=quantile_numerator,
            quantile_denominator=quantile_denominator,
            name="MagnitudePercentageRatio",
            feets_feature=feets_feature,
            feets_skip_test="feets uses different quantile type",
        ),
    )


FluxPercentileRatioMid20 = generate_test_magnitude_percentile_ratio(0.40, 0.05, "FluxPercentileRatioMid20")
FluxPercentileRatioMid50 = generate_test_magnitude_percentile_ratio(0.25, 0.05, "FluxPercentileRatioMid50")
FluxPercentileRatioMid80 = generate_test_magnitude_percentile_ratio(0.10, 0.05, "FluxPercentileRatioMid80")


class TestMaximumSlope(_Test):
    name = "MaximumSlope"

    feets_feature = "MaxSlope"

    def naive(self, t, m, sigma):
        return np.max(np.abs((m[1:] - m[:-1]) / (t[1:] - t[:-1])))


class TestMean(_Test):
    name = "Mean"

    feets_feature = "Mean"

    def naive(self, t, m, sigma):
        return np.mean(m)


class TestMeanVariance(_Test):
    name = "MeanVariance"

    feets_feature = "Meanvariance"
    feets_skip_test = "feets uses biased statistics"

    def naive(self, t, m, sigma):
        return np.std(m, ddof=1) / np.mean(m)


class TestMedian(_Test):
    name = "Median"

    def naive(self, t, m, sigma):
        return np.median(m)


class TestMedianAbsoluteDeviation(_Test):
    name = "MedianAbsoluteDeviation"

    feets_feature = "MedianAbsDev"


class TestMedianBufferRangePercentage(_Test):
    # feets says it uses 0.1 of amplitude (a half range between max and min),
    # but factually it uses 0.1 of full range between max and min
    quantile = 0.2

    name = "MedianBufferRangePercentage"
    kwargs = dict(quantile=quantile)

    feets_feature = "MedianBRP"


class TestPercentAmplitude(_Test):
    name = "PercentAmplitude"

    feets_feature = "PercentAmplitude"
    feets_skip_test = "feets divides value by median"

    def naive(self, t, m, sigma):
        median = np.median(m)
        return max(np.max(m) - median, median - np.min(m))


class TestPercentDifferenceMagnitudePercentile(_Test):
    quantile = 0.05

    name = "PercentDifferenceMagnitudePercentile"
    kwargs = dict(quantile=quantile)

    feets_feature = "PercentDifferenceFluxPercentile"
    feets_skip_test = "feets uses different quantile type"


class TestReducedChi2(_Test):
    name = "ReducedChi2"

    def naive(self, t, m, sigma):
        w = 1.0 / np.square(sigma)
        return np.sum(np.square(m - np.average(m, weights=w)) * w) / (m.size - 1)


class TestRoms(_Test):
    name = "Roms"


class TestSkew(_Test):
    name = "Skew"

    feets_feature = "Skew"
    feets_skip_test = "feets uses biased statistics"

    def naive(self, t, m, sigma):
        return stats.skew(m, bias=False)


class TestStandardDeviation(_Test):
    name = "StandardDeviation"

    feets_feature = "Std"
    feets_skip_test = "feets uses biased statistics"

    def naive(self, t, m, sigma):
        return np.std(m, ddof=1)


class TestStetsonK(_Test):
    name = "StetsonK"

    feets_feature = "StetsonK"

    def naive(self, t, m, sigma):
        x = (m - np.average(m, weights=1.0 / sigma**2)) / sigma
        return np.sum(np.abs(x)) / np.sqrt(np.sum(np.square(x)) * m.size)


class TestWeightedMean(_Test):
    name = "WeightedMean"

    def naive(self, t, m, sigma):
        return np.average(m, weights=1.0 / sigma**2)


class TestAllPy(_Test):
    # Most of the features are for mags
    phot_types = frozenset(["mag"])

    def setup_method(self):
        features = []
        py_features = []
        for cls in _Test.__subclasses__():
            if cls.name is None:
                continue

            try:
                py_features.append(getattr(lc_py, cls.name)(**cls.kwargs))
            except AttributeError:
                continue
            features.append(getattr(lc_ext, cls.name)(**cls.kwargs))
        self.rust = lc_ext.Extractor(*features)
        self.py_feature = lc_py.Extractor(*py_features)


class TestBenchmarkParallel(_Test):
    phot_types = frozenset(["mag"])
    features_subset = [
        "AndersonDarlingNormal",
        "Beyond1Std",
        "Cusum",
        "EtaE",
        "InterPercentileRange",
        "Kurtosis",
        "LinearTrend",
        "MaximumSlope",
        "Mean",
        "MeanVariance",
        "MedianAbsoluteDeviation",
        "MedianBufferRangePercentage",
        "PercentAmplitude",
        "PercentDifferenceMagnitudePercentile",
        "Skew",
        "StandardDeviation",
        "StetsonK",
    ]

    def setup_method(self):
        features = []
        py_features = []
        for cls in _Test.__subclasses__():
            if cls.name is None and cls.name not in self.features_subset:
                continue

            try:
                py_features.append(getattr(lc_py, cls.name)(**cls.kwargs))
            except AttributeError:
                continue
            features.append(getattr(lc_ext, cls.name)(**cls.kwargs))
        self.rust = lc_ext.Extractor(*features)
        self.py_feature = lc_py.Extractor(*py_features)

    def rust_many(self, n_jobs, light_curves):
        return self.rust.many(
            light_curves,
            n_jobs=int(n_jobs),
            sorted=True,
            check=False,
        )

    def py_parallel(self, n_jobs, light_curves):
        return Parallel(n_jobs=int(n_jobs))(delayed(self.py_feature)(*lc) for lc in light_curves)

    def rust_parallel(self, n_jobs, light_curves):
        return Parallel(n_jobs=int(n_jobs))(delayed(self.rust)(*lc) for lc in light_curves)

    @pytest.mark.parametrize("n_core", np.linspace(1, 8, 8))
    @pytest.mark.multi
    def test_benchmark_many_rust(self, benchmark, n_core):
        light_curves = []

        for _ in range(1000):
            t, m, sigma = self.random_data()
            light_curves.append((t, m, sigma))

        benchmark.group = f"{n_core}_cores"
        benchmark.name = "rust multiprocessing (many)"

        benchmark(self.rust_many, n_core, light_curves)

    @pytest.mark.parametrize("n_core", np.linspace(1, 8, 8))
    @pytest.mark.multi
    def test_benchmark_multi_python(self, benchmark, n_core):
        light_curves = []

        for _ in range(1000):
            t, m, sigma = self.random_data()
            light_curves.append((t, m, sigma))

        benchmark.group = f"{n_core}_cores"
        benchmark.name = "python multiprocessing"

        benchmark(self.py_parallel, n_core, light_curves)

    @pytest.mark.parametrize("n_core", np.linspace(1, 8, 8))
    @pytest.mark.multi
    def test_benchmark_multi_rust(self, benchmark, n_core):
        light_curves = []

        for _ in range(1000):
            t, m, sigma = self.random_data()
            light_curves.append((t, m, sigma))

        benchmark.group = f"{n_core}_cores"
        benchmark.name = "rust multiprocessing"

        benchmark(self.rust_parallel, n_core, light_curves)


class TestAllNaive(_Test):
    # Most of the features are for mags
    phot_types = frozenset(["mag"])

    def setup_method(self):
        features = []
        self.naive_features = []
        for cls in _Test.__subclasses__():
            if cls.naive is None or cls.name is None:
                continue
            if not cls.add_to_all_features:
                continue
            if not cls.add_to_all_features:
                continue
            features.append(getattr(lc_ext, cls.name)(*cls.args, **cls.kwargs))
            self.naive_features.append(cls().naive)
        self.rust = lc_ext.Extractor(*features)

    def naive(self, t, m, sigma):
        return np.concatenate([np.atleast_1d(f(t, m, sigma)) for f in self.naive_features])


class TestAllFeets(_Test):
    feets_skip_test = "skip for TestAllFeets"

    # Most of the features are for mags
    phot_types = frozenset(["mag"])

    def setup_method(self):
        features = []
        feets_features = []
        for cls in _Test.__subclasses__():
            if cls.feets_feature is None or cls.name is None:
                continue
            if not cls.add_to_all_features:
                continue
            if not cls.add_to_all_features:
                continue
            features.append(getattr(lc_ext, cls.name)(*(cls.args + tuple(cls.kwargs.values()))))
            feets_features.append(cls.feets_feature)
        self.rust = lc_ext.Extractor(*features)
        self.feets_extractor = feets.FeatureSpace(only=feets_features, data=["time", "magnitude", "error"])



from dataclasses import dataclass

from ..dataclass_field import dataclass_field
from ._base import BaseSingleBandFeature


@dataclass()
class MagnitudePercentageRatio(BaseSingleBandFeature):
    quantile_numerator: float = dataclass_field(default=0.4, kw_only=True)
    quantile_denominator: float = dataclass_field(default=0.05, kw_only=True)

    def _eval_single_band(self, t, m, sigma=None):
        try:
            from scipy.stats.mstats import mquantiles
        except ImportError:
            raise ImportError("scipy is required for MagnitudePercentageRatio feature, please install it")

        n1, n2 = mquantiles(m, [self.quantile_numerator, 1 - self.quantile_numerator], alphap=0.5, betap=0.5)
        d1, d2 = mquantiles(m, [self.quantile_denominator, 1 - self.quantile_denominator], alphap=0.5, betap=0.5)
        return (n2 - n1) / (d2 - d1)

    @property
    def size_single_band(self):
        return 1


__all__ = ("MagnitudePercentageRatio",)



"""Re-implementation of iminuit.cost.LeastSquares with an arbitrary data format"""

from typing import Callable, Dict, Tuple

import numpy as np

try:
    from iminuit import Minuit
except ImportError:
    LeastSquares = None
else:
    # Following https://iminuit.readthedocs.io/en/stable/notebooks/generic_least_squares.html
    class LeastSquares:
        errordef = Minuit.LEAST_SQUARES

        def __init__(self, model: Callable, parameters: Dict[str, Tuple[float, float]], *, x, y, yerror):
            self.model = model
            self.x = x
            self.y = y
            self.yerror = yerror
            self._parameters = parameters

        @property
        def ndata(self):
            return len(self.y)

        def __call__(self, *par):
            ym = self.model(self.x, *par)
            return np.sum(np.square((self.y - ym) / self.yerror))


__all__ = ["LeastSquares"]



from dataclasses import dataclass

import numpy as np

from ..dataclass_field import dataclass_field
from ._base import BaseSingleBandFeature


@dataclass()
class MedianBufferRangePercentage(BaseSingleBandFeature):
    quantile: float = dataclass_field(default=0.1, kw_only=True)

    def _eval_single_band(self, t, m, sigma=None):
        median = np.median(m)
        return np.count_nonzero(np.abs(median - m) < self.quantile * (np.max(m) - np.min(m)) / 2) / len(m)

    @property
    def size_single_band(self):
        return 1


__all__ = ("MedianBufferRangePercentage",)



import numpy as np
from numpy.testing import assert_allclose

from light_curve.light_curve_py import Kurtosis


def test_kurtosis():
    m = [
        1.52403507,
        -0.47585435,
        1.30247288,
        -0.26769695,
        -1.89584998,
        0.30886293,
        -1.08824993,
        1.17991399,
        -0.61845487,
        0.12171829,
    ]
    feature = Kurtosis()
    desired = -0.704411
    actual = feature(np.linspace(0, 1, len(m)), m, None)
    assert_allclose(actual, desired)



import numpy as np
from numpy.testing import assert_allclose

from light_curve.light_curve_py import AndersonDarlingNormal


def test_adnormal():
    m = np.arange(0, 9)
    feature = AndersonDarlingNormal()
    desired = feature(m, m, None)
    actual = 0.155339690
    assert_allclose(actual, desired)



import numpy as np
from numpy.testing import assert_allclose

from light_curve.light_curve_py import OtsuSplit


def test_otsu_1():
    feature = OtsuSplit()
    m = np.array([0.5, 1.5, 1.5, 1.5])
    t = np.linspace(0, 1.7, 4)
    actual = feature(t, m, None)
    desired = [1.0, 0.0, 0.0, 0.25]
    assert_allclose(actual, desired)


def test_otsu_2():
    feature = OtsuSplit()
    m = np.array([0.45, 0.5, 0.52])
    t = np.linspace(0, 1.7, 3)
    actual = feature(t, m, None)
    desired = [0.06, 0.0, np.std([0.5, 0.52], ddof=1), 0.33333]
    assert_allclose(actual, desired, rtol=1e-04)


def test_otsu_thr():
    feature = OtsuSplit()
    m = np.array([0.45, 0.5, 0.52])
    actual = feature.threshold(m)
    desired = 0.5
    assert_allclose(actual, desired, rtol=1e-04)


def test_otsu_thr_sort():
    m = np.array([0.45, 0.5, 0.52])
    m_shuffled = np.array([0.52, 0.45, 0.5])
    actual = OtsuSplit.threshold(m)
    desired = OtsuSplit.threshold(m_shuffled)
    assert_allclose(actual, desired, rtol=1e-04)


def test_otsu_sort():
    feature = OtsuSplit()
    m = np.array([0.45, 0.5, 0.52])
    m_shuffled = np.array([0.52, 0.45, 0.5])
    t = np.linspace(0, 1.7, 3)
    actual = feature(t, m, None)
    desired = feature(t, m_shuffled, None)
    assert_allclose(actual, desired, rtol=1e-04)



import pytest

from light_curve.light_curve_py import ColorOfMedian


def test_color_of_median():
    m = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]
    band = ["g", "g", "g", "r", "r", "r", "i"]

    feature = ColorOfMedian("g", "r")
    actual = feature(t=m, m=m, sigma=None, band=band)
    desired = -3.0
    assert actual == desired


def test_color_of_median_wrong_number_of_args():
    with pytest.raises(TypeError):
        _feature = ColorOfMedian()

    with pytest.raises(TypeError):
        _feature = ColorOfMedian("gr")

    with pytest.raises(TypeError):
        _feature = ColorOfMedian("g", "r", "i")



# Import all Python features
from .light_curve_py import *

# Hide Python features with Rust equivalents
from .light_curve_ext import *

# Hide Rust Extractor with universal Python Extractor
from .light_curve_py import Extractor

from .light_curve_ext import __version__



import numpy as np

from ._base import BaseSingleBandFeature


class MedianAbsoluteDeviation(BaseSingleBandFeature):
    def _eval_single_band(self, t, m, sigma=None):
        return np.median((np.abs(m - np.median(m))))

    @property
    def size_single_band(self):
        return 1


__all__ = ("MedianAbsoluteDeviation",)



import numpy as np
from numpy.testing import assert_allclose

from light_curve.light_curve_py import Roms


def test_roms_const_data():
    feature = Roms()
    n = 100
    t = np.arange(n)
    m = np.ones_like(t)
    sigma = np.ones_like(t)
    actual = feature(t, m, sigma)
    desired = 0.0
    assert_allclose(actual, desired)


def test_roms_periodic_data():
    feature = Roms()
    n = 100
    t = np.linspace(0, 2 * np.pi, n)
    m = 2 * np.sin(t)
    sigma = np.ones_like(t)
    actual = feature(t, m, sigma)
    desired = 4 / np.pi
    assert_allclose(actual, desired, rtol=3 / np.sqrt(n))


def test_roms_norm_data_1():
    rng = np.random.default_rng(0)
    n = 100
    t = np.linspace(0, 1, n)
    m = rng.normal(0, 1, n)
    sigma = np.ones_like(t)
    feature = Roms()
    actual = feature(t, m, sigma)
    desired = 2 / np.sqrt((2 * np.pi))
    assert_allclose(actual, desired, rtol=3 / np.sqrt(n))


def test_roms_norm_data_2():
    rng = np.random.default_rng(0)
    n = 10000
    t = np.linspace(0, 1, n)
    m = rng.normal(0, 1, n)
    sigma = np.ones_like(t)
    feature = Roms()
    actual = feature(t, m, sigma)
    desired = 2 / np.sqrt((2 * np.pi))
    assert_allclose(actual, desired, rtol=3 / np.sqrt(n))


def test_roms_expon_data_1():
    rng = np.random.default_rng(0)
    n = 100
    t = np.linspace(0, 1, n)
    m = rng.exponential(2, n)
    sigma = np.ones_like(t)
    feature = Roms()
    actual = feature(t, m, sigma)
    desired = 2 * np.log(2)
    assert_allclose(actual, desired, rtol=3 / np.sqrt(n))


def test_roms_expon_data_2():
    rng = np.random.default_rng(0)
    n = 10000
    t = np.linspace(0, 1, n)
    m = rng.exponential(2, n)
    sigma = np.ones_like(t)
    feature = Roms()
    actual = feature(t, m, sigma)
    desired = 2 * np.log(2)
    assert_allclose(actual, desired, rtol=3 / np.sqrt(n))


def test_roms_gamma_data_1():
    rng = np.random.default_rng(0)
    n = 100
    t = np.linspace(0, 1, n)
    m = rng.gamma(2, 1, n)
    sigma = np.ones_like(t)
    feature = Roms()
    actual = feature(t, m, sigma)
    desired = 1.0518265193
    assert_allclose(actual, desired, rtol=3 / np.sqrt(n))


def test_roms_gamma_data_2():
    rng = np.random.default_rng(0)
    n = 10000
    t = np.linspace(0, 1, n)
    m = rng.gamma(2, 1, n)
    sigma = np.ones_like(t)
    feature = Roms()
    actual = feature(t, m, sigma)
    desired = 1.0518265193
    assert_allclose(actual, desired, rtol=3 / np.sqrt(n))



from .light_curve import *


# A hack to bypass the lack of package support in PyO3
# https://github.com/PyO3/pyo3/issues/1517#issuecomment-808664021
def __register_submodules():
    import sys

    sys.modules["light_curve.light_curve_ext.ln_prior"] = ln_prior


__register_submodules()


def feature_from_json(s: str) -> JSONDeserializedFeature:
    """Deserialize a JSON string into a Feature object."""
    return JSONDeserializedFeature(s)



import numpy as np
import scipy.stats as st
from numpy.testing import assert_allclose

from light_curve.light_curve_py import PeakToPeakVar


def test_ptpvar():
    feature = PeakToPeakVar()
    rng = np.random.default_rng(0)
    n = 100
    lmd = 200
    t = np.arange(n)
    flux = rng.poisson(lmd, n)
    sigma = np.sqrt(flux)

    a1 = st.norm.ppf(q=(0.025) ** (1 / n), loc=n - np.sqrt(n), scale=np.sqrt(n + 0.5**2))
    b1 = st.norm.ppf(q=(0.975) ** (1 / n), loc=n - np.sqrt(n), scale=np.sqrt(n + 0.5**2))
    a2 = st.norm.ppf(q=((((0.025 - 1) * (-1)) ** (1 / n)) - 1) * (-1), loc=n + np.sqrt(n), scale=np.sqrt(n + 0.5**2))
    b2 = st.norm.ppf(q=((((0.975 - 1) * (-1)) ** (1 / n)) - 1) * (-1), loc=n + np.sqrt(n), scale=np.sqrt(n + 0.5**2))

    a = (a1 - b2) / (a1 + a2)
    b = (b1 - a2) / (b1 + b2)
    c = (a + b) / 2
    d = (b - a) / 2

    actual = feature(t, flux, sigma)
    desired = c
    atol = d
    assert_allclose(actual, desired, atol=atol)



import numpy as np
from numpy.testing import assert_allclose

from light_curve.light_curve_py import InterPercentileRange


def test_intpercrange_1():
    m = np.arange(1.0, 7.0)
    feature = InterPercentileRange()
    actual = feature(m, m)
    desired = 3.0
    assert_allclose(actual, desired)


def test_intpercrange_2():
    m = np.arange(1.0, 10.0)
    feature = InterPercentileRange(quantile=0.5)
    actual = feature(m, m)
    desired = 0.0
    assert_allclose(actual, desired)



from numpy.testing import assert_allclose

from light_curve.light_curve_py import MaximumSlope


def test_maxslope():
    m = [2, 4, 15, 3]
    t = [1, 3, 4, 8]
    feature = MaximumSlope()
    actual = feature(t, m, None)
    desired = 11.0
    print(actual)
    assert_allclose(actual, desired)



from enum import IntEnum
from typing import Iterable, List

import numpy as np

from light_curve.light_curve_py.features.rainbow._bands import Bands

__all__ = ["create_parameters_class"]


def baseline_parameter_name(band: str) -> str:
    return f"baseline_{band}"


def baseline_band_name(name: str) -> str:
    if name.startswith("baseline_"):
        return name[len("baseline_") :]

    return None


def create_int_enum(cls_name: str, attributes: Iterable[str]):
    return IntEnum(cls_name, {attr: i for i, attr in enumerate(attributes)})


def create_parameters_class(
    cls_name: str,
    *,
    common: List[str],
    bol: List[str],
    temp: List[str],
    bands: Bands,
    with_baseline: bool,
):
    """Create an IntEnum class for RainbowFit parameters

    Parameters
    ----------
    cls_name : str
        Name of the class to create
    common : list of str
        Common parameters for both bolometric and temperature models
    bol : list of str
        Bolometric model parameters, without common parameters
    temp : list of str
        Temperature model parameters, without common parameters
    bands : list of str
        Unique list of bands in the dataset. It is used to generate baseline
        parameters when `with_baseline` is True.
    with_baseline : bool
        Whether to include baseline parameters, one per band in `bands`.
    """
    attributes = common + bol + temp
    if with_baseline:
        baseline = list(map(baseline_parameter_name, bands.names))
        attributes += baseline

    enum = create_int_enum(cls_name, attributes)

    enum.all_common = common
    enum.common_idx = np.array([enum[attr] for attr in common])

    enum.bol = bol
    enum.bol_idx = np.array([enum[attr] for attr in enum.bol])
    enum.all_bol = common + bol
    enum.all_bol_idx = np.array([enum[attr] for attr in enum.all_bol])

    enum.temp = temp
    enum.temp_idx = np.array([enum[attr] for attr in enum.temp])
    enum.all_temp = common + temp
    enum.all_temp_idx = np.array([enum[attr] for attr in enum.all_temp])

    enum.with_baseline = with_baseline
    if with_baseline:
        enum.all_baseline = baseline
        enum.baseline_idx = np.array([enum[attr] for attr in enum.all_baseline])
        enum.baseline_parameter_name = staticmethod(baseline_parameter_name)
        enum.baseline_band_name = staticmethod(baseline_band_name)

        band_idx_to_baseline_idx = {
            band_idx: enum[baseline_parameter_name(band_name)] for band_idx, band_name in zip(bands.index, bands.names)
        }
        enum.lookup_baseline_idx_with_band_idx = np.vectorize(band_idx_to_baseline_idx.get, otypes=[int])

    return enum



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, Sequence

import numpy as np

from light_curve.light_curve_py.dataclass_field import dataclass_field
from light_curve.light_curve_py.warnings import mark_experimental


@dataclass
class BaseMultiBandFeature(ABC):
    @property
    @abstractmethod
    def size(self) -> int:
        pass

    @property
    @abstractmethod
    def is_band_required(self) -> bool:
        pass

    @property
    @abstractmethod
    def is_multiband_supported(self) -> bool:
        pass

    def _eval(self, *, t, m, sigma, band):
        """It could be missed if _eval_and_fill is re-implemented"""
        raise NotImplementedError("_eval is missed")

    @abstractmethod
    def _eval_and_fill(self, *, t, m, sigma, band, fill_value):
        """It has a default implementation, but it requires _eval to be implemented"""
        try:
            a = self._eval(t=t, m=m, sigma=sigma, band=band)
            if np.any(~np.isfinite(a)):
                raise ValueError
            return a
        except (ValueError, ZeroDivisionError, RuntimeError) as e:
            if fill_value is not None:
                return np.full(self.size, fill_value)
            raise e

    @mark_experimental
    def __post_init__(self) -> None:
        pass

    def _normalize_input(self, *, t, m, sigma, band, sorted, check):
        t = np.asarray(t)
        m = np.asarray(m)

        if sigma is not None:
            sigma = np.asarray(sigma)

        if band is None and self.is_band_required:
            raise ValueError("band is required")
        if band is not None:
            if not self.is_multiband_supported:
                raise ValueError(
                    "(band != None) is not supported by this feature instance, consider to pass a band array "
                    "or recreate the feature instance with different parameters"
                )
            band = np.asarray(band)
            if band.ndim != 1:
                raise ValueError("band must be None or 1D array-like")

        if check:
            if np.any(~np.isfinite(t)):
                raise ValueError("t values must be finite")
            if np.any(~np.isfinite(m)):
                raise ValueError("m values must be finite")
            if sigma is not None and np.any(np.isnan(sigma)):
                raise ValueError("sigma must have no NaNs")

        if sorted is None:
            diff = np.diff(t)
            if np.any(diff == 0):
                raise ValueError("t must be unique")
            if np.any(diff < 0):
                raise ValueError("t must be sorted")
        elif not sorted:
            idx = np.argsort(t)
            t = t[idx]
            m = m[idx]
            if sigma is not None:
                sigma = sigma[idx]
            if band is not None:
                band = band[idx]

        return t, m, sigma, band

    def __call__(self, t, m, sigma=None, band=None, *, sorted=None, check=True, fill_value=None):
        t, m, sigma, band = self._normalize_input(t=t, m=m, sigma=sigma, band=band, sorted=sorted, check=check)
        return self._eval_and_fill(t=t, m=m, sigma=sigma, band=band, fill_value=fill_value)

    def many(self, lcs, *, sorted=None, check=True, fill_value=None, n_jobs=-1):
        """Extract features in bulk

        This exists for computability only and doesn't support parallel
        execution, that's why `n_jobs=1` must be used
        """
        if n_jobs != 1:
            raise NotImplementedError("Parallel execution is not supported by this feature, use n_jobs=1")
        return np.stack([self(*lc, sorted=sorted, check=check, fill_value=fill_value) for lc in lcs])


@dataclass
class BaseSingleBandFeature(BaseMultiBandFeature):
    bands: Optional[Sequence[str]] = dataclass_field(default=None, kw_only=True)

    @property
    @abstractmethod
    def size_single_band(self) -> int:
        pass

    @abstractmethod
    def _eval_single_band(self, *, t, m, sigma):
        pass

    @property
    def n_bands(self) -> int:
        if self.bands is None:
            return 1
        return len(self.bands)

    @property
    def size(self) -> int:
        return self.n_bands * self.size_single_band

    @property
    def is_band_required(self) -> bool:
        return self.bands is not None

    @property
    def is_multiband_supported(self) -> bool:
        return self.bands is not None

    def _eval_and_fill_single_band(self, *, t, m, sigma, fill_value):
        try:
            a = self._eval_single_band(t=t, m=m, sigma=sigma)
            if np.any(~np.isfinite(a)):
                raise ValueError
            return a
        except (ValueError, ZeroDivisionError, RuntimeError) as e:
            if fill_value is not None:
                return np.full(self.size_single_band, fill_value)
            raise e

    def _eval_and_fill(self, *, t, m, sigma, band, fill_value):
        if self.bands is None:
            return self._eval_and_fill_single_band(t=t, m=m, sigma=sigma, fill_value=fill_value)

        values = []
        for band_to_calc in self.bands:
            band_mask = band == band_to_calc
            t_band = t[band_mask]
            m_band = m[band_mask]
            if sigma is None:
                sigma_band = None
            else:
                sigma_band = sigma[band_mask]
            v = self._eval_and_fill_single_band(t=t_band, m=m_band, sigma=sigma_band, fill_value=fill_value)
            values.append(np.atleast_1d(v))

        return np.concatenate(values)



import numpy as np
from numpy.testing import assert_allclose

from light_curve.light_curve_py import Eta


def test_eta():
    a = np.array([1, 2, 3, 3, 3, 4, 4])
    feature = Eta()
    actual = feature(np.linspace(0, 1, len(a)), a, None)
    desired = 0.43750000000000006
    assert_allclose(actual, desired)



import numpy as np
from numpy.testing import assert_allclose

from light_curve.light_curve_py import PercentDifferenceMagnitudePercentile


def test_pdiffmperc():
    m = np.arange(1.0, 7.0)
    feature = PercentDifferenceMagnitudePercentile(quantile=0.25)
    actual = feature(m, m)
    desired = 3.0 / 3.5
    assert_allclose(actual, desired)



import numpy as np

from ._base import BaseSingleBandFeature


class Eta(BaseSingleBandFeature):
    def _eval_single_band(self, t, m, sigma=None):
        n = len(m)
        m_std = np.var(m, ddof=1)
        m_sum = np.sum((m[1:] - m[:-1]) ** 2)
        return m_sum / ((n - 1) * m_std)

    @property
    def size_single_band(self):
        return 2


__all__ = ("Eta",)



import numpy as np

from ._base import BaseSingleBandFeature


class EtaE(BaseSingleBandFeature):
    def _eval_single_band(self, t, m, sigma=None):
        n = len(m)
        m_std = np.var(m, ddof=1)
        m_sum = np.sum(((m[1:] - m[:-1]) / (t[1:] - t[:-1])) ** 2)
        return m_sum * (t[n - 1] - t[0]) ** 2 / ((n - 1) ** 3 * m_std)

    @property
    def size_single_band(self):
        return 1


__all__ = ("EtaE",)



import numpy as np
from numpy.testing import assert_allclose
from scipy import signal

from light_curve.light_curve_py import StetsonK


def test_stetsonk_1():
    feature = StetsonK()
    x = np.linspace(0.0, 2 * np.pi, 2000)
    sin = np.sin(x)
    error = np.ones(2000)
    actual = feature(x, sin, error)
    desired = np.sqrt(8) / np.pi
    assert_allclose(actual, desired, rtol=1e-03)


def test_stetsonk_2():
    feature = StetsonK()
    x = np.linspace(0.0, 1.0, 1000)
    sawtooth = signal.sawtooth(2 * np.pi * 5 * x)
    error = np.ones(1000)
    actual = feature(x, sawtooth, error)
    desired = np.sqrt(12) / 4
    assert_allclose(actual, desired, rtol=1e-03)
